#!/opt/bin/gawk -f
#

#
# This BEGIN block is for initialization only.
#


BEGIN {
	configdir = "/opt/etc/yawk";
	bindir = "/opt/lib/yawk";
	yawkconf = "/opt/etc/yawk/yawk.conf";
	configdelim = ",";
	readparameters();
	}

function readparameters() {
	filename = ENVIRON["YAWK_CONF"];
	if (filename == "")
		filename = yawkconf;

	while (getline <filename > 0) {
		if ($1 == "configdir")
			configdir = $2;
		else if ($1 == "bindir")
			bindir = $2;
		else if ($1 == "configdelim")
			configdelim = $2;
		else if ($1 == "cgidir")
			cgidir = $2;
		}
		
	close (filename);
	return (0);
	}



function skipws(line) {
	sub(/^[ \t\r\n]+/, "", line);
	return (line);
	}

function noctrl(line) {
	sub(/[ \t\r\n]+$/, "", line);
	return (line);
	}

function quote(string) {
	gsub(/'/, "'\\''", string);
	string = "'" string "'";
	return (string);
	}

function unslash(path, rfe) {
	gsub(/\/+/, "/", path);
	if (rfe != 0)
		sub(/\/+$/, "", path);

	return (path);
	}

function getbasename(fn) {
	sub(/^.*\//, "", fn);
	return (fn);
	}

function getdirname(fn) {
	if (fn !~ /\//)
		return ("");

	sub(/\/[^\/]*$/, "", fn);
	return (fn);
	}


function cgidecode(str,   k, x, result) {
	hex = "123456789ABCDEF";
	result = "";
	gsub(/+/, " ", str);
	while ((k = index(str, "%")) > 0) {
		result = result substr(str, 1, k-1);
		str = substr(str, k+1);
		x = index(hex, substr(str, 1, 1)) * 16 + index(hex, substr(str, 2, 1));
		result = result sprintf ("%c", x);
		str = substr(str, 3);
		}

	result = result str;
	return (result);
	}

function getcgivars(querystring,   i, k, n, x, name, val) {
	delete CGIVAR;
	n = split(querystring, x, "&");
	for (i=1; i<=n; i++) {
		if ((k = index(x[i], "=")) > 0) {
			name = "CGI_" toupper(substr(x[i], 1, k-1));
			val = cgidecode(substr(x[i], k+1));
			if ((name in CGIVAR)  &&  ENVIRON[name] != "")
				ENVIRON[name] = ENVIRON[name] ", " val;
			else
				ENVIRON[name] = val;

			CGIVAR[name] = 1;
			}
		}

	return (0);
	}

function rmvar(var,   p) {
	p = "(^|&)" var "=[^&]*";
	gsub(p, "", ENVIRON["QUERY_STRING"]);

	return (0);
	}

function getval(var, a1, a2,   v) {
	v = a2;
	if (var in ENVIRON  &&  (v = ENVIRON[var]) != "")
		;
	else if (a1 != "")
		v = a1;

	return (v);
	}

function makeoptions(arr, opt, val,   string, i, n, x, q) {
	q = "\001";
	if (substr(opt, 1, 1) == "-") {
		opt = substr(opt, 2);
		if (opt in arr)
			delete arr[opt];
		}
	else if (opt != "") {
		q = arr[opt];
		arr[opt] = val;
		}

	n = split("m,w,s,o,b,c,fs,abs", x, ",");
	for (i=1; i <= n; i++) {
		if (x[i] in arr  &&  arr[x[i]] != ""  &&  arr[x[i]] != "-")
			string = string configdelim x[i] "=" arr[x[i]];
		}

	if (q != "\001")
		arr[opt] = q;

	return (string);
	}

function makebaseurl(arr, opt, val,   string, baseurl) {
	string = makeoptions(arr, opt, val);
	if (string != ""  ||  config != "")
		string = "-" config string "/";

	baseurl = ENVIRON["SCRIPT_NAME"] "/" string;
	return (baseurl);
	}


function eatone(n, r,   i, val) {
	val = "";
	for (i=1; i<=n; i++) {
		val = val " " $i;
		$i = "";
		}

	$0 = skipws($0);
	if (r != 0)
		return ($0);

	return (skipws(val));
	}

function rewritehref(href, type,   p) {
	if (USESTDIO != "yes")
		return (href);

	if (type == "css")
		p = (substr(href, 1, 1) == "/")? HREFBASE href: href;
	else if (type == "wiki") {
		if (href == "..")
			p = href;
		else if (href ~ /\.yawk$/)
			p = href ".html";
		else {
#			while (match(href, /[[:lower:]][[:upper:]]/) > 0) {
#				p = p substr(href, 1, RSTART) "-";
#				href = substr(href, RSTART+1);
#				}
#
#			p = tolower(p href ".html");
			p = tolower(gensub(/([[:lower:]])([[:upper:]])/, "\\1-\\2", "g", href)) ".html";
			}
		}

	return (p);
	}

function __computestat(path, sbuf, line,   i, n, x, y, z) {
	sbuf[path, "type"] = ".";

#printf ("cs= %s\n", line) >>STDERR;
	if ((n = split(line, x, /[ \t]+/)) == 9)
		;	# ok - "old style" ls
	else if (n == 8  &&  x[6] ~ /-[0-9][0-9]-/)
		;	# also ok - "modern style"
	else {
		return (1);
		}

	sbuf[path, "type"] = substr(x[1], 1, 1) == "d"? "dir": "file";
	gsub(/\/+/, "/", x[n]);
	sbuf[path, "filename"] = x[n];
	sbuf[path, "size"] = x[5];

	if (n == 8) {
		split(x[6], z, "-");
		split(x[7], y, ":");
		line = sprintf ("%04d %02d %02d %02d %02d %02d",
				z[1], z[2], z[3],
				y[1], y[2], y[3]);
		}
	else {
		n = split("jan feb mar apr may jun jul aug sep oct nov dec", y, " ");
		for (i=1; i <= n; i++)
			z[y[i]] = sprintf ("%d", i);

		if (x[8] !~ /:/)
			line = sprintf ("%04d %02d %02d 00 00 00", x[8], z[tolower(x[6])], x[7]);
		else {
			split(x[8], y, ":");
			line = sprintf ("%04d %02d %02d %02d %02d %02d",
					strftime("%Y", systime()), z[tolower(x[6])], x[7],
					y[1], y[2], y[3]);
			}
		}

	sbuf[path, "rawtime"] = line;
	sbuf[path, "mtime"] = mktime(line);
	sbuf[path, "lastmod"] = strftime("%a, %d %b %Y %H:%M:%S %z", sbuf[path, "mtime"]);

	return (0);
	}

function stat(path, sbuf,   cmd, line, rc) {
	cmd = sprintf ("/bin/ls -fld %s 2>/dev/null", quote(path));
	cmd | getline line;
	close (cmd);

	rc = __computestat(path, sbuf, line);
	return (rc);
	}


function sendheader(status, content, nocache) {
	printf ("Status: %s\n", status);
	printf ("Content-Type: %s\n", content);

	if (nocache == 1  ||  wikimode == "sidebar"  ||  wikimode == "sidewiki")
		printf ("Pragma: no-cache\n");

	printf ("\n");

	respstatus = status;
	respcontent = content;

	return (0);
	}

function send404resp(path) {
	sendheader(status, "text/html", 1);
	printf ("<B>404 File not found</B><P>\n");
	printf ("File does not exits: %s<BR>\n", path);
	printf ("You may try %s?op=edit to create the file.<BR>\n", BASEADR path);

	return (0);
	}

function sendredirect(location, error, cookie) {
	location = unslash(location, 0);

	printf ("Status: %s\n", 302);
	printf ("Content-Type: %s\n", "text/html");
	printf ("Location: %s\n", location);

	printf ("\n");

	printf ("<A HREF=\"%s\">%s</A>\n", location, location);
	printf ("\n");

	respstatus = 302;
	respcontent = location;

	return (0);
	}

function makeurl(adr, modifier) {
	if (modifier != "") {
		if (index(modifier, "://") > 0)
			adr = modifier;
		else if (substr(modifier, 1, 1) == "?")
			adr = adr modifier;
		else if (substr(modifier, 1, 1) == "/")
			adr = BASEADR modifier;		# SCRIPTBASE modifier;
		else {
			sub(/[\/]+$/, "", adr);
#			adr = BASEADR modifier;
			adr = adr "/" modifier;
			}
		}

	return (adr);
	}

function sendredir2(adr, alt) {
	adr = makeurl(adr, alt);
	sendredirect(adr);

	return (0);
	}


function getmimetype(filename,   ext, mime, type) {
	sub(/^.*\//, "", filename);
	if (filename ~ CamelCase)
		return ("text/yawk");
	else if (match(filename, /^[-a-z0-9]+$/))
		return ("text/plain");
	else if (match(filename, /\.[^.]+(\.gz|\.Z)?$/) > 0) {
		if ((ext = substr(filename, RSTART)) == ".yawk")
			return ("text/yawk");

		if (1  ||  mimetype["text/html"] == "") {
			mimetype["text/html"] = " .html .htm ";
			mimetype["image/jpeg"] = " .jpg .jpeg ";
			mimetype["image/gif"] = " .gif ";
			mimetype["image/png"] = " .png ";
			mimetype["text/plain"] = " .txt .css ";
			}

		ext = " " ext " ";
		mime = "";

		for (type in usertype) {
			if (index(usertype[type], ext) > 0) {
				mime = type;
				break;
				}
			}

		if (mime == "") {
			for (type in mimetype) {
				if (index(mimetype[type], ext) > 0) {
					mime = type;
					break;
					}
				}
			}

		if (mime == "")
			mime = "application/octet-stream";

		return (mime);
		}

	return ("text/yawk");
	}


function rewritename(filename, checkspecial,   i, k, n, x, y) {
	gsub(/\/+/, "/", filename);
	n = split(filename, x, /\/+/);
	k = 0;
	for (i=1; i<=n; i++) {
		if (x[i] == "") {
			if (i == n)
				y[k++] = "";

			continue;
			}

		if (x[i] == ".")
			continue;
		else if (x[i] == "..") {
			if (k >= 1)
				k--;
			}
		else
			y[k++] = x[i];
		}

	filename = y[0];
	for (i=1; i<k; i++)
		filename = filename "/" y[i];

	if (checkspecial == 1) {
		if (match(filename, /^(.*\/)?(\$[^/]+)$/, x) > 0) {
			if (x[2] == "$auto-create")
				filename = sprintf ("%s%s-%04X.yawk", x[1], ENVIRON["SERIAL"], ENVIRON["PID"]);
			}
		}

	gsub(/\.+/, ".", filename);
	gsub(/[^-\/\.\+_0-9A-Za-z]+/, "", filename);

	gsub(/\/\.+/, "", filename);
	gsub(/^\.+/, "", filename);

	return (filename);
	}

function makefn(name,   fn) {
	if (name == "")
		return ("");

	if (substr(name, 1, 1) == "/")
		fn = wikipath "/" rewritename(name);
	else
		fn = wikidir "/" rewritename(name);

	return (fn);
	}

function _xx_mksubdirlocation(path, app, lastname, lastapp, tg,   s, i, n, p, x) {
	s = sprintf ("<A HREF=\"%s%s\"%s>%s</A>", BASEADR, app, tg, word["top"]);
#printf ("path= %s, app= %s, lastname= %s, lastapp= %s\n", path, app, lastname, lastapp) >>STDERR;
	if (path != "/") {
		n = split(path, x, "/");
		while (n > 1  &&  x[n] == "")
			n--;

		p = "";
		if (lastname == "")
			lastname = x[n];

		for (i=1; i<=n; i++) {
			if (x[i] == "")
				continue;

			p = p x[i] "/";
			s = s sprintf(" &gt; <A HREF=\"%s%s\">%s</A>",
						BASEADR p,
						(i != n)? app: lastapp,
						(i != n)? x[i]: lastname);
			}
		}

	return (s);
	}

function mksubdirlocation(base, path, app, lastname, lastapp, tg,   s, i, n, p, x) {
	if (base == "")
		base = BASEADR;

	n = 0;
	if (path != "/")
		n = split(path, x, "/");

	while (n > 0  &&  x[n] == "")
		n--;

	p = "";
	if (lastname == "")
		lastname = (n > 0)? x[n]: word["top"];

#printf ("path= %s, n= %d, app= %s, lastname= %s, lastapp= %s\n", path, n, app, lastname, lastapp) >>STDERR;
	for (i=0; i<=n; i++) {
		if (i == 0) {
			s = sprintf ("<A HREF=\"%s%s\"%s>%s</A>", base,
					(n != 0)? app: lastapp, tg,
					(n != 0)? word["top"]: lastname);
			continue;
			}

		if (x[i] == "")
			continue;

		p = p x[i] "/";
		s = s sprintf(" &gt; <A HREF=\"%s%s\"%s>%s</A>",
					base p,
					(i != n)? app: lastapp, tg,
					(i != n)? x[i]: lastname);
		}

#printf ("s= %s\n", s) >>STDERR;
	return (s);
	}

function getdisplayval(val,   x) {
	split(val, x, /[ ,.]+/);
	if ((p = ENVIRON["CGI_P"]+0) > 0)
		x[1] = p;

	if ((p = ENVIRON["CGI_Q"]+0) > 0)
		x[2] = p;

	if (x[1] < 1)
		x[1] = 3;

	if (x[2] < 1)
		x[2] = 120;

	val = sprintf ("%d,%d", x[1], x[2]);
	return (val);
	}

function __gettext(n, list, b,   i, cmd, mime, p, x) {
	if (b == "")
		b = getdisplayval(ENVIRON["CGI_B"]);

#	cmd = sprintf ("%s -bb %s", wikiparser, getdisplayval(ENVIRON["CGI_B"]));
	cmd = sprintf ("%s -bb %s", wikiparser, b);
	printf ("") |& cmd;
	for (i=1; i<=n; i++) {
#printf (">> i= %d, line= %s\n", i, list[i]) >>STDERR;
		split(list[i], x, "\t");
		if (x[1] == ""  ||  x[1] ~ /[:?]/)
			continue;

		if (x[1] ~ /\/$/) {
			line = getbasename(substr(x[1], 1, length(x[1]) - 1)) "/";
			mime = "dir";
			}
		else if ((mime = getmimetype(x[1])) != "text/yawk")
			line = getbasename(x[1]);
		else {
			printf ("%s\n", x[1]) |& cmd;
			fflush(cmd);

			cmd |& getline line;
			sub(/^[^\t]*\t/, "", line);
			}

		list[i] = x[1] "\t" line "\t" x[2] "\t" mime;
#printf (">> i= %d, line= %s\n", i, list[i]) >>STDERR;
		}

	close (cmd);
	return (0);
	}


function insertheader(header,   filename, line) {
	if (header == "")
		return (0);

	if (substr(header, 1, 2) == "./")
		filename = wikidir "/" substr(header, 2);
	else if (substr(header, 1, 1) == "/")
		filename = wikipath header;
	else
		filename = wikidir "/" header;

	while (getline line <filename > 0)
		printf ("%s\n", line);

	close (filename);

	return (0);
	}

function makeadr(base, href, path,   p) {
	if (base == "")
		base = BASEADR;

	gsub(/\$/, "/" path, href);

	if (href ~ /^[a-z]+:/)
		p = href;
	else if (substr(href, 1, 1) == "/")
		p = base substr(href, 2);
	else if (substr(href, 1, 2) == "./")
		p = substr(href, 3);
	else
		p = href;

	return (p);
	}

function subnav(fn, nav, tg,   e, i, n, p, q, r, targetseq, bn, x, y) {
	bn = fn;  sub(/^.*\//, "", bn);
	delete nav;
	n = split(rellist " navigator:sub:_parent linker:sub:_parent compound:sub", x, " ");
	for (i=1; i<=n; i++) {
		split(x[i], y, ":");
		e = y[1];
		p = y[2];

		if (! (e in fileconf)) {
			if (p == "pnx"  &&  isSmall == 0  &&  isPlain == 0)
				nav["non"] = nav["non"] "<SPAN CLASS=\"nav-disabled\">" word[e] "</SPAN>\n";

			continue;
			}

		targetseq = (y[3] != "")? sprintf(" TARGET=\"%s\"", y[3]): tg;
		if (ak[e] != "")
			targetseq = targetseq " ACCESSKEY=\"" ak[e] "\"";

		if (nav[p] != "")
			nav[p] = nav[p] " | ";

		r = sprintf("<A HREF=\"%s\"%s TITLE=\"%s\">%s</A>", makeadr("", fileconf[e]), targetseq, e, word[e]);
		if (fileconf[e] == ""  ||  fileconf[e] == bn)
			nav[p] = nav[p] word[e] "\n";
		else
			nav[p] = nav[p] r "\n";

		if (fileconf[e] == "")
			nav["non"] = nav["non"] "<SPAN CLASS=\"nav-disabled\">" word[e] "</SPAN>\n";
		else {
			nav["non"] = nav["non"] "<SPAN CLASS=\"nav-enabled\">" r "</SPAN>\n";
			q++;
			}
		}

	if (q == 0)
		nav["non"] = "";

	return (0);
	}


function mkshortcuts(file, tg,   i, n, p, q, u, x, y) {
	p = "R,$|X";
	n = split(p, x, "|");

	for (i=(ENVIRON["CGI_OP"] != "")? n-1: 1; i<=n; i++) {
		split(x[i], y, ",");

#printf ("y= %s, file= %s\n", y[1], file) >>STDERR;
		u = y[2];
		if (match(u, /\$/) > 0) {
			if (file == "")
				continue;

			gsub(/$/, file, u);
			}

		if (u ~ /:\/\//)
			;		# nothing
		else if (substr(u, 1, 1) == "/")
			u = BASEADR substr(u, 2);
		else
			u = BASEADR wikisubdir u;

		if (y[1] == "R")
			q = q sprintf ("<A HREF=\"%s\"%s>R</A>", file, tg);
		else if (y[1] == "X") {
			if (option["w"] == "text") {
				q = q sprintf ("<A HREF=\"%s\" TARGET=\"_parent\">X</A>",
						makebaseurl(option, "w", "") path);
				}
			}
		}

	return (q);
	}

function _xx_mkshortcuts(url,   p) {
	if (ENVIRON["CGI_OP"] != "")
		return ("");

	p = sprintf ("<A HREF=\"%s\">R</A>", url);
	if (option["w"] == "text") {
		p = p sprintf ("<A HREF=\"%s\" TARGET=\"_parent\">X</A>",
				makebaseurl(option, "w", "") path);
		}

	if (p != "")
		p = " | " p;

	return (p);
	}


function getnavstring(string) {
	if (uppercasenav == 1)
		return (toupper(string));

	return (string);
	}


function getnav(base, file, top, nav, tg,   i, n, p, q, s, x, y) {
	if (base == "")
		base = BASEADR;

	p = "index,,?op=main|list,,?op=list|edit,,+?op=edit|info,,+?op=info"\
	    "|search,1,?op=search|pnx,1|top,0,#TOP";

	n = split(p, x, "|");
	for (i=1; i<=n; i++) {
		split(x[i], y, ",");
		if (y[2] != ""  &&  y[2] != top)
			continue;
		else if (y[1] == "edit"  &&  (readonly != 0  ||  op == "edit"))
			continue;
		else if (y[1] == "search"  &&  op == "search")
			continue;
		else if (y[1] == "pnx") {
#printf ("pnx= %s\n", nav["pnx"]) >>STDERR;
			if (nav["pnx"] == ""  ||  isPlain != 0  ||  isPDA != 0)
				continue;
			}


		if (y[1] == "top") {
			q = q sprintf ("| <A HREF=\"%s\"%s ACCESSKEY=\"%s\">%s</A>\n", y[3], tg, ak[y[1]], word[y[1]]);
			continue;
			}
		else if (y[1] == "pnx") {
			q = q sprintf ("| <A HREF=\"%s\">%s</A>\n",
					"javascript:show('pnxnav', '')", word["pnx"]);
			continue;
			}

		p = base wikisubdir;
		if (substr(y[3], 1, 1) != "+")
			p = p y[3];
		else if (file == "")
			continue;
		else
			p = p file substr(y[3], 2);

		s = "";
		if (ak[y[1]] != "")
			s = sprintf (" ACCESSKEY=\"%s\" TITLE=\"%s\"", ak[y[1]], y[1]);


		if (q != "")
			q = q " | ";

		q = q sprintf ("<A HREF=\"%s\"%s%s>%s</A>\n",
				p, tg, s, word[y[1]]);

		if (y[1] == "edit"  &&  side != "") {
			q = q sprintf ("<A HREF=\"%s&target=%s\"%s ACCESSKEY=\"*\">%s</A>\n",
				p, otherside, tg, "*");
			}

		if (wikimode == "crawler"  &&  y[1] == "list")
			break;
		}

	return (q);
	}

function navigator(base, url, top,   i, c, p, s, r, nav, file, line, tg) {
	if (applyfirefoxpatch != 0)
		tg = " TARGET=\"_self\"";
	else if (option["w"] == "list")
		tg = " TARGET=\"_parent\"";

	base = BASEADR wikisubdir;
#printf ("base= %s\n", base) >>STDERR;
	if (op == "printer"  ||  op == "save-as")
		return (0);

	if (wikimode == "raw")
		return (0);

#	sub(/.*\//, "", url);
	file = getbasename(url);
	if (file != ""  &&  top == 1  &&  showtopnav == 0)
		return (0);

	if (top != 0) {
		printf ("<A NAME=\"TOP\"></A>\n");
		subnav(file, nav, "");

#printf ("pnx= %s\n", nav["pnx"]) >>STDERR;
		}
	else {
		printf ("<P>\n");
		printf ("<A NAME=\"BOTTOM\"></A>\n");
		printf ("</DIV>\n");
		}


	if (top == 0  &&  option["w"] != "list"  &&  (op == ""  ||  op == "search"))
		insertheader(wikifooter);

	if (top != 0  &&  option["w"] != "list"  &&  (op == ""  ||  op == "search"))
		insertheader(wikiheader);

	if (usetablenav == 0)
		printf ("<DIV CLASS=\"%s\">\n", (top == 1)? "top-nav": "bottom-nav");
	else {
		printf ("<TABLE VALIGN=\"TOP\" CLASS=\"%s-nav\" WIDTH=\"100%\" CELLSPACING=\"0\" CELLPADDING=\"0\">\n", top == 0? "bottom": "top");
		printf ("<TR>\n");
		}

# ???	if (showdirnav == 1  &&  wikimode == "") {
	if (showdirnav == 1) {
		if (usetablenav != 0)
			printf ("<TD CLASS=\"nav\" ALIGN=\"LEFT\">\n");

#printf ("file= %s, wikiindex= %s\n", file, wikiindex) >>STDERR;
		p = makebaseurl(option, "w", "");
		if (file == wikiindex)
			p = mksubdirlocation(p, wikisubdir, "", "...", "?op=browse", tg);
		else
			p = mksubdirlocation(p, wikisubdir, "", "", "", tg);

		printf ("%s", p);
		if (usetablenav != 0)
			printf ("</TD>\n");
		else
			printf (" :: ");
		}

	if (usetablenav != 0)
		printf ("<TD CLASS=\"nav\" ALIGN=\"RIGHT\">\n");

	p = makebaseurl(option, "w", "");
	p = getnav(p, file, top, nav, tg);
	printf ("%s\n", p);
	if ((p = mkshortcuts(file)) != "")
		printf ("| %s\n", p);

	if (usetablenav != 0)
		printf ("</TD></TR></TABLE>\n");


	if (top == 0)
		printf ("</DIV>\n");
	else if (top == 1) {
		printf ("</DIV>\n");

		if (nav["pnx"] != ""  ||  nav["sub"] != "") {
			printf ("<DIV CLASS=\"subnav\" ALIGN=\"RIGHT\">\n");
			if (isPDA == 1  ||  isPlain == 1)
				;
			else if (nav["pnx"] != "") {
				printf ("<SPAN ID=\"pnxnav\" CLASS=\"subnav\" STYLE=\"display: none;\">%s</SPAN>\n", nav["pnx"]);
				if (nav["sub"] != "")
					printf (" &#160; ");
				}

			if (nav["sub"] != "")
				printf ("<SPAN CLASS=\"subnav\">%s</SPAN>\n", nav["sub"]);

			printf ("&#160;");
			printf ("</DIV>\n");
			}

#		printf ("<P>\n");
		if (op != "") {
			printf ("<P>\n");
			printf ("<DIV CLASS=\"content-%s\">\n", op);
			}
		else
			printf ("<DIV CLASS=\"content\">\n");

		if (wikimode == "crawler") {
			printf ("<CENTER>Your browser is \"%s\" which has been identified as webcrawler.\n", ENVIRON["HTTP_USER_AGENT"]);
			printf ("This is why this page is specifically rendered.</CENTER><P>\n");
			}
		}

	return (0);
	}


function _xx_navigator(base, url, top,   i, c, p, s, r, nav, tg) {
	if (applyfirefoxpatch != 0)
		tg = " TARGET=\"_self\"";

	base = BASEADR wikisubdir;
#printf ("base= %s\n", base) >>STDERR;
	if (op == "printer"  ||  op == "save-as")
		return (0);

	if (wikimode == "raw")
		return (0);

#	sub(/.*\//, "", url);
	url = getbasename(url);
	if (url != ""  &&  top == 1  &&  showtopnav == 0)
		return (0);

	if (top != 0) {
		printf ("<A NAME=\"TOP\"></A>\n");
		subnav(url, nav, "");
#printf ("pnx= %s\n", nav["pnx"]) >>STDERR;
		}
	else {
		printf ("<P>\n");
		printf ("<A NAME=\"BOTTOM\"></A>\n");
		printf ("</DIV>\n");
		}


	if (usetablenav == 0)
		printf ("<DIV CLASS=\"%s\">\n", (top == 1)? "top-nav": "bottom-nav");
	else {
		printf ("<TABLE VALIGN=\"TOP\" CLASS=\"%s-nav\" WIDTH=\"100%\" CELLSPACING=\"0\" CELLPADDING=\"0\">\n", top == 0? "bottom": "top");
		printf ("<TR>\n");
		}

# ???	if (showdirnav == 1  &&  wikimode == "") {
	if (showdirnav == 1) {
		if (usetablenav != 0)
			printf ("<TD CLASS=\"nav\" ALIGN=\"LEFT\">\n");

#printf ("url= %s, wikiindex= %s\n", url, wikiindex) >>STDERR;
		if (url == wikiindex)
			p = mksubdirlocation("", wikisubdir, "", "...", "?op=browse", "");
		else
			p = mksubdirlocation("", wikisubdir, "", "", "", "");

		printf ("%s", p);
		if (usetablenav != 0)
			printf ("</TD>\n");
		else
			printf (" :: ");
		}

	if (usetablenav != 0)
		printf ("<TD CLASS=\"nav\" ALIGN=\"RIGHT\">\n");

	printf ("<A HREF=\"%s?op=main\" ACCESSKEY=\"%s\">%s</A>\n", BASEADR wikisubdir, ak["index"], word["index"]);
	printf ("| <A HREF=\"%s?op=list\" ACCESSKEY=\"%s\">%s</A>\n", BASEADR wikisubdir, ak["list"], word["list"]);

#	if (showdirnav != 0) {
#		if ((p = "contents") in fileconf  &&  url != fileconf[p])
#			printf("<A HREF=\"%s\"%s>%s</A> | ", makeadr("", fileconf[p]), tg, word[p]);
#		}
#	else
#		printf ("<A HREF=\"%s?op=main\" ACCESSKEY=\"%s\">%s</A>\n", BASEADR wikisubdir, ak["index"], word["index"]);
#		
#	printf ("<A HREF=\"%s?op=list\" ACCESSKEY=\"%s\">%s</A>\n", BASEADR wikisubdir, ak["list"], word["list"]);

#printf ("mode= %s\n", wikimode) >>STDERR;
	if (ENVIRON["CGI_OP"] != "")
		;	
	else if (wikimode == "crawler")
		;	# do nothing
	else if (url != "") {
		if (readonly == 0) {
			printf ("| <A HREF=\"%s?op=edit\" ACCESSKEY=\"%s\">%s</A>\n", url, ak["edit"], word["edit"]);
			if (side != "")
				printf ("<A HREF=\"%s?op=edit&target=%s\" ACCESSKEY=\"*\">%s</A>\n", url, otherside, "*");
			}

		printf ("| <A HREF=\"%s?op=info\" ACCESSKEY=\"%s\">%s</A>\n", url, ak["info"], word["info"]);

#		if (side != "") {
#			if (wikiopt["lockframe"] == "unlocked")
#				printf ("| <A HREF=\"%s%s%s\" ACCESSKEY=\"%s\">%s</A>\n", lockurl, wikisubdir, path, ak["lock"], word["lock"]);
#			else if (wikiopt["lockframe"] == "locked")
#				printf ("| <A HREF=\"%s%s%s\" ACCESSKEY=\"%s\">%s</A>\n", unlockurl, wikisubdir, path, ak["unlock"], word["unlock"]);
#			}

		if (top == 1  ||  showtopnav == 0) {
			printf ("| <A HREF=\"%s\" ACCESSKEY=\"%s\">%s</A>\n", BASEADR wikisubdir "?op=search", ak["search"], word["search"]);

				if (showsearch == 2  &&  rellinks["search"] != "") {
					printf ("| Search: <FORM CLASS=\"navsearch\" ACTION=\"%s\">\n", BASEADR);
					printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"search\">\n");
					printf ("<INPUT CLASS=\"navsearch\" TYPE=\"TEXT\" ACCESSKEY=\"S\" NAME=\"pattern\" VALUE=\"\">\n");
					printf ("</FORM>\n");
					}
			}
 
 		if (top == 1) {
			if (nav["pnx"] != ""  &&  isPlain == 0  &&  isPDA == 0) {
				printf ("| <A HREF=\"%s\">%s</A>\n",
						"javascript:show('pnxnav', '')", word["pnx"]);
				}

#			printf ("| <A HREF=\"%s\">R</A>", url);
#
#			if (option["w"] == "text") {
#				printf ("<A HREF=\"%s\" TARGET=\"_parent\">X</A>",
#						makebaseurl(option, "w", "") path);
#				}

			printf ("%s", mkshortcuts(url));
			printf ("\n");
			}

		if (top == 0  ||  showtopnav == 0)
			printf ("| <A HREF=\"#TOP\" ACCESSKEY=\"T\">%s</A>\n", getnavstring("Top"));
		}
	else {
		if (op != "info")
			printf ("| <A HREF=\"%s?op=info\" ACCESSKEY=\"F\">%s</A>\n", url, getnavstring("Info"));

		if (readonly == 0)
			printf ("| <A HREF=\"%s?op=new\" ACCESSKEY=\"N\">%s</A>\n", base, getnavstring("New"));

		printf ("| <A HREF=\"%s?op=search\" ACCESSKEY=\"S\">%s</A>\n", base, getnavstring("Search"));

		if (option["w"] == "text") {
			printf ("<A HREF=\"%s\" TARGET=\"_parent\">X</A>",
					makebaseurl(option, "w", "") path);
			}

		printf ("\n");
		}

	if (usetablenav != 0)
		printf ("</TD></TR></TABLE>\n");


	if (top == 0)
		printf ("</DIV>\n");
	else if (top == 1) {
		printf ("</DIV>\n");

		if (nav["pnx"] != ""  ||  nav["sub"] != "") {
			printf ("<DIV CLASS=\"subnav\" ALIGN=\"RIGHT\">\n");
			if (isPDA == 1  ||  isPlain == 1)
				;
			else if (nav["pnx"] != "") {
				printf ("<SPAN ID=\"pnxnav\" CLASS=\"subnav\" STYLE=\"display: none;\">%s</SPAN>\n", nav["pnx"]);
				if (nav["sub"] != "")
					printf (" &#160; ");
				}

			if (nav["sub"] != "")
				printf ("<SPAN CLASS=\"subnav\">%s</SPAN>\n", nav["sub"]);

			printf ("&#160;");
			printf ("</DIV>\n");
			}

#		printf ("<P>\n");
		if (op != "") {
			printf ("<P>\n");
			printf ("<DIV CLASS=\"content-%s\">\n", op);
			}
		else
			printf ("<DIV CLASS=\"content\">\n");

		if (wikimode == "crawler") {
			printf ("<CENTER>Your browser is \"%s\" which has been identified as webcrawler.\n", ENVIRON["HTTP_USER_AGENT"]);
			printf ("This page is why this page is specifically rendered.</CENTER><P>\n");
			}
		}

	return (0);
	}



function endhtml() {
	printf ("&#160;<P>\n");
	printf ("</BODY>\n");
	printf ("</HTML>\n");
	return (0);
	}

function _xx_insertrelated(rel, ctype, path, byval,   line) {
	if (byval == 0)
		printf ("  <LINK REL=\"%s\" TYPE=\"%s\" HREF=\"%s\">\n", rel, ctype, path);
	else {
		printf ("  <%s type=\"%s\">\n", rel, ctype);
		while (getline line <path > 0)
			printf ("    %s\n", line);

		close (path);
		printf ("  </%s>\n", rel);
		}

	return (0);
	}

function starthtml(title, css, keepopen, currentfile,
				i, k, p, type, line, n, x) {
	if (op == "printer"  &&  printercss != "")
		css = printercss;

	printf ("<HTML>\n");
	printf ("<HEAD>\n");
	printf ("  <TITLE>%s</TITLE>\n", title);
	printf ("  <X-INFO PATH-INFO=\"%s\">\n", ENVIRON["PATH_INFO"]);

	printf ("  <META HTTP-EQUIV=\"Content-Type\" VALUE=\"text/html; charset=%s\">\n", charset);
	if ((filename SUBSEP "mtime") in sbuf)
		printf ("  <META HTTP-EQUIV=\"Last-Modified\" CONTENT=\"%s\">\n", strftime("%a, %d %b %Y %H:%M:%S %z", sbuf[filename, "mtime"]));

	if (ENVIRON["QUERY_STRING"] != ""  &&  ENVIRON["CGI_OP"] != "list")
		printf ("  <META NAME=\"robots\" CONTENT=\"noindex\">\n");

#	if (ENVIRON["QUERY_STRING"] != ""  &&  ENVIRON["CGI_OP"] != "list")
#		printf ("  <META NAME=\"robots\" CONTENT=\"noindex, nofollow\">\n");
#	else if (wikimode != ""  ||  side != "")
#		printf ("  <META NAME=\"robots\" CONTENT=\"noindex, nofollow\">\n");


	if (internalstyle != 0) {
		twoem  = wikimode == "pda"? ".3em": "2em";
		halfem = wikimode == "pda"? ".2em": ".5em";

		if ((i = option["fs"]+0) > 0)
			fontsize = i;
		else if (wikimode == "sidebar"  ||  wikimode == "pda")
			fontsize = 9;
		else
			fontsize = 11;

		printf ("  <STYLE TYPE=\"text/css\">\n");
		printf ("    BODY { color: #0003; background: #FFFFFF; font-family: %s; font-size: %dpt; }\n", fontfamily, fontsize);
		printf ("    P, DIV, TD, SELECT, INPUT { font-family: %s; font-size: %dpt; }\n", fontfamily, fontsize);
		printf ("    P.sectionheader { font-size: 110%%; } \n");
		printf ("    H3 { margin-bottom: .3em; margin-top: 1em; }\n");
		printf ("    UL, OL { padding-left: %s; }\n", twoem);
		printf ("    A:link, A:visited { text-decoration: none; }\n");
		printf ("    A:hover { color: #FF0000; }\n");
		printf ("    PRE { font-family: \"courier new\", monospace; margin-left: 1em; margin-right: 3em; }\n");

		printf ("    DIV.example { background: #F0F0F0; margin-bottom: .5em; border-top: 1px solid black; border-bottom: 1px solid black; }\n");
		# printf ("    PRE.example { background: #F0F0F0; margin: 0; padding: .5em; }\n");
		printf ("    PRE.example { background: #F0F0F0; margin: 0; padding: .5em; border-top: 1px solid black; border-bottom: 1px solid black; overflow: auto; }\n");
		
		printf ("    DIV.example-title { font-size: .8em; text-align: center; background: #F0F0F0; margin: 0 .3em 0 .3em; border-top: 1px solid black; }\n");

		printf ("    BODY.nav { padding: 0; margin: 0; text-align: right; background: #FFF8F8; }\n");
		printf ("\n");
		printf ("    DIV.xx-searchform {\n");
		printf ("	right: 3em; top: 2em; font-size: 1em;\n");
		printf ("	padding: .2em;\n");
		printf ("	}\n");
		printf ("    DIV.searchform { padding: .2em; }\n");
		printf ("    DIV.openform { padding: .2em; }\n");
		printf ("    DIV.sl2 { align: center; font-size: .9em; padding: .4em 0 .4em 0; margin: .4em 0 .4em 0; border-bottom: 1px solid black }\n");
		printf ("    OL.search { font-size: %s; }\n",
				option["o"] != "du"  &&  isSmall == 0? listfontsize: "100%");
		printf ("\n");
		printf ("    DIV.table { }\n");
		printf ("    TABLE { margin-left: 1em; }\n");
		printf ("    CAPTION.table { vertical-align: bottom; font-size: .8em; text-align: center; background: #F0F0F0; border-top: 1px solid black; }\n");
		printf ("    TD { color: #003; font-size: %dpt; vertical-align: top; }\n", fontsize);
		printf ("\n");
		printf ("    TABLE.browser { margin-left: 0em; padding-left: .5em; width: 100%; }\n");
		printf ("    TABLE.browser TD { padding: .2em .4em .2em .4em; }\n", fontsize);
		printf ("    A.browser { text-decoration: none; color: black; }\n");
		printf ("    A.browser:hover { color: red; }\n");
		printf ("\n");
		printf ("    BODY.text, BODY.dir { padding: 0; margin: 0; background: #FFFFFF; }\n");
		printf ("    TABLE.top-nav, TABLE.bottom-nav {\n");
		printf ("		margin-left: 0em; border-top: 1px solid blue; border-bottom: 1px solid blue;\n");
		printf ("		padding: 0 2em 0 .5em; background: rgb(255, 255, 204);\n");
		printf ("		height: 1.3em; width: 100%; font-size: 90%; font-weight: bold; margin-top: -2px;\n");
		printf ("		}\n");
		printf ("    TABLE.bottom-nav { position: %s; bottom: -1px; }\n", isPDA != 0? "block": "fixed");
		printf ("    TD.nav { };\n");
		printf ("\n");
		printf ("    DIV.top-nav, DIV.bottom-nav { text-align: right; padding: 0 10px 0 10px; margin: 0; }\n");
		printf ("    DIV.top-nav { border-bottom: 1px solid black; background: #F8F8FF; }\n");
		printf ("    DIV.bottom-nav { border-top: 1px solid red; }\n");
		printf ("\n");
		printf ("    DIV.content, DIV.content-search { padding: %s; }\n", halfem);
		printf ("    TABLE.dir TD { padding: .2em .4em .2em .4em; }\n", fontsize);
		printf ("\n");
		printf ("    DIV.pnx-fixed {\n");
		printf ("	position: %s; width: 100%; background: #F8F8F8; border-bottom: 1px solid #808080;\n", (isPDA == 0  &&  isPlain == 0)? "fixed": "static");
		printf ("	text-align: left; font-size: 90%; padding: 0px 10px 0px 10px; margin: 0;\n");
		printf ("	}\n");
		printf ("    DIV.pnx-fixed A { color: black; }\n");
		printf ("    SPAN.nav-disabled { color: #666666; }\n");
		printf ("    SPAN.nav-enabled { padding: 0 1px 0 1px; }\n");
		printf ("    SPAN.nav-enabled:hover { padding: 0 2px 0 2px; background: #E0E0E0; color: black; }\n");
		printf ("\n");
		printf ("    DIV.subnav { text-align: right; margin: 5px 5px -1em 5px; }\n");
		printf ("    SPAN.subnav { border: 1px solid black; background: #F8F8F8; text-align: right; x-font-variant: small-caps; padding: 0px 10px 0px 10px; x-margin: 5px; }\n");
		printf ("    DIV.abs { font-size: .9em; border: 1px solid blue; margin: .3em %s .3em %s; padding: .3em; display: %s; }\n",
					twoem, twoem, option["abs"] == "hide"? "none": "block");
		printf ("\n");
		printf ("    BODY.list { padding: 0; margin: 0; background: #F8FFF8; }\n");
		printf ("\n");
		printf ("    TEXTAREA.edit { width: 95%; height: 85%; font-size: %dpt; }\n", fontsize);
		printf ("    DL { margin-left: 1em; }\n");
		printf ("    DD { margin-left: 1em; margin-bottom: .3em; }\n");
		printf ("\n");
		printf ("    DIV.linklist, DIV.news {\n");
		printf ("      margin-left: %s; margin-bottom: 1em; margin-top: 0em; width: 20em;\n", twoem);
		printf ("      float: right; padding: 0.2em; font-size: 80%; background-color: #F8F8F8; border: 1px solid blue;\n");
		printf ("      }\n");
		printf ("    UL.link { line-height: 1.2em; padding-left: 1.5em; margin-left: 0em; }\n");
		printf ("    DIV.news { padding: .5em; }\n");
		printf ("\n");
		printf ("    DIV.nl-title { margin: .2em; font-weight: bold; }\n");
		printf ("    DIV.nl-text { margin: .2em 0 .2em 1.2em; }\n");
		printf ("\n");
		printf ("    SPAN.c { x-font-weight: bold; font-family: \"courier new\", monospace; color: #0000FF; }\n");
		printf ("    SPAN.i { font-weight: bold; color: #0000FF; }\n");
		printf ("    DIV.wikiseq { background: #F0F0F0; margin: 0 %s 0 %s; padding: .5em; border: 1px dashed black; }\n", twoem, twoem);
		printf ("    DIV.notice { margin-left: %s; padding: .4em; margin-right: %s; border: 1px solid red; }\n", twoem, twoem);
		printf ("    DIV.important { margin-left: %s; padding: .4em; margin-right: %s; background: #FF3333; }\n", twoem, twoem);
		printf ("\n");
		printf ("    DIV.compound { padding: .5em .5em .5em .5em; }\n");
		printf ("    DIV.compound DIV.abs { display: none; }\n");
		printf ("    DIV.linklist DIV.abs { display: none; }\n");
		printf ("    DIV.compound H2 { border-top: 1px dashed grey; }\n");

		if (nstyle > 0) {
			printf ("\n");
			for (i=1; i<=nstyle; i++)
				printf ("    %s\n", cstyle[i]);
			}

		printf ("  </STYLE>\n");
		}

	for (i=1; i<=csscount; i++)
		printf ("  <LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"%s\">\n", stylesheet[i]);

	for (i=1; i<=scripts; i++)
		printf ("  <SCRIPT TYPE=\"text/javascript\" SRC=\"%s\"></SCRIPT>\n", scriptfile[i]);

	printf ("<SCRIPT TYPE=\"text/javascript\">\n");
	printf ("function show(id, val) {\n");
	printf ("	document.getElementById(id).style.display = val;\n");
	printf ("	if (val == 'block')\n");
	printf ("		document.search.s.focus();\n");
	printf ("	}\n");
	printf ("\n");
	printf ("function settarget(osurl) {\n");
	printf ("	if (document.search.nw.checked)\n");
	printf ("		document.search.target = '_blank';\n");
	printf ("	else if (document.search.o.value == 'os'  ||  document.search.o.value == 'sw') {\n");
	printf ("		document.search.target = '_self';\n");
	printf ("		document.search.action = osurl;\n");
	printf ("		}\n");
	printf ("	return (true);\n");
	printf ("	}\n");
	printf ("</SCRIPT>\n");

	printf ("<STYLE TYPE=\"text/css\">\n");
	printf ("  TABLE.sc { margin: 0; align: right; }\n");
	printf ("  TD.search { font-size: 1em; x-vertical-align: middle; }\n");
	printf ("  TD.search  SELECT { font-size: 1em; x-vertical-align: middle; }\n");
	printf ("</STYLE>\n");

	printf ("</HEAD>\n");
	printf ("\n");

	if (css != " "  &&  keepopen == 0) {
		line = "";
		for (i=1; i<=jshc; i++)
			line = line " " bodyhandler[i];

		if (wikimode == "raw")
			printf ("<BODY CLASS=\"raw\">\n");
		else
			printf ("<BODY%s>\n", line);
		}

	return (0);
	}

function escape(line, quote) {
	gsub(/&/, "\\&amp;", line);
	gsub(/>/, "\\&gt;", line);
	gsub(/</, "\\&lt;", line);

	if (quote != 0)
		gsub(/"/, "\\&quot;", line);

	return (line);
	}



function _xx_subst(string,   c, v, v2, w, x, y, searchhref) {
	searchhref = ("_SEARCH_" in var)? var["_SEARCH_"]: "?op=search";
	x = string;
	y = "";
	while (match(x, /(\$\([A-Za-z_][.A-Za-z0-9_]*\))|(\$\([^':\)]*[:][\?!]))/) > 0) {
		y = y substr(x, 1, RSTART-1);
		v = substr(x, RSTART+2, RLENGTH-3);
		x = substr(x, RSTART+RLENGTH);

		if (match(v, /:/) == 0) {
			if (tolower(substr(v, 1, 5)) == "attr_"  &&
			    (v2 = tolower(substr(v, 6))) in filedata) {
				w = filedata[v2];
				}
			else if (v in var)
				w = var[v];
			else if (tolower(v) in filedata)
				w = filedata[tolower(v)];
			else
				w = ENVIRON[v];

			y = y w;
			}
		else {
			c = substr(v, RSTART+1, 1);
			v = substr(v, 1, RSTART-1);
			gsub(/ /, "+", v);
			gsub(/,/, "%2C", v);
			if (c == "?"  ||  userelsearch != 0)
				w = sprintf ("%s&pattern=%s", searchhref, v);
			else if (c == "!")
				w = sprintf ("%s&pattern=(^|\\W)%s(\\W|$)", searchhref, v);

			y = y w;
			}
		}

	string = y x;
	return (string)
	}



function readrelfile(relfile, fn,   p, k) {
#printf ("relfile= %s\n", relfile) >>STDERR
	while (getline <relfile > 0) {
		if (match($0, /^:wc:[ \t]+rel[ \t]+/) == 0)
			continue;

		$0 = noctrl(substr($0, RLENGTH+1));
#printf ("line= %s\n", $0) >>STDERR;
		if ($1 ~ /:$/) {
			p = substr($1, 1, length($1) - 1);
			fileconf[p] = eatone(1, 1);
			}
		else {
#printf ("rellink= %s, fn= %s\n", $0, fn) >>STDERR;
			rellink[++nrel] = $0;
			if ($0 == fn)
				k = nrel;
			}
		}

	close (relfile);
	if (nrel > 0) {
		fileconf["first"] = rellink[1];
		if (nrel > 1)
			fileconf["last"] = rellink[nrel];

		if (k != 0) {
			fileconf["prev"] = (k > 1)? rellink[k-1]: "";
			fileconf["next"] = (k < nrel)? rellink[k+1]: "";
			}
		}

#printf ("nrel= %d, prev= %s\n", nrel, fileconf["prev"]) >>STDERR;
	return (0);
	}

function readfile(filename, n, text, relp, clear,   line, path, p, opt, par, x,
			fn, relfile) {
	if (clear != 0)
		delete fileconf;

	delete rellink;
	if (relp != 0)
		path = sprintf ("%s/%s", wikipath, filename)
	else {
		path = filename;
		filename = substr(filename, length(wikipath) + 2);
		}

#printf ("filename= %s, path= %s\n", filename, path) >>STDERR;
	while (getline line <path > 0) {
		if (match(line, /^(:wc:)[ \t]+([^ \t\r]+)([ \t\r]+([^\r]+)?)?[ \t\r]*$/, x) > 0) {
			opt = x[2];
			par = (x[4] == "")? filename: x[4];

			if (x[2] == "isnav"  ||  x[2] == "nav"  ||  x[2] == "navigator") {
				if (par !~ /[:?]/) {
					p = sprintf ("?op=search&o=%s&ph=1&type=n&expr=%s",
							(isSmall != 0)? "du": "lr", "%3D" par);
					}

				fileconf["navigator"] = p;
				}
			else if (x[2] == "linker") {
				if (par !~ /[:?]/) {
					p = sprintf ("?op=search&o=%s&ph=1&type=l&expr=%s",
							(isSmall != 0)? "du": "lr", par);
					}

				fileconf["linker"] = p;
				}
			else if (x[2] == "compound") {
				if (par !~ /[:?]/)
					p = sprintf ("?op=compound&s=%s", par);

				fileconf["compound"] = p;
				}
			else if (opt == "relfile") {
				if (substr(par, 1, 1) == "/")
					relfile = basedir par;
				else
					relfile = getdirname(path) "/" par;
				}
			else if (opt == "isindex")
				fileconf[opt] = fileconf[opt] " " par;
			else
				fileconf[opt] = par;

#			continue;
			}

		text[++n] = line;
		}

	close (path);

	if (relfile != "") {
#		p = filename;
#		sub(/^.*\//, "", p);
		p = getbasename(filename);
		readrelfile(relfile, p);
		}

	if ("search" in fileconf  &&  fileconf["search"] == ".")
		fileconf["search"] = (isPDA == 1  ||  isPlain == 1)? "?op=search": "javascript:show('searchform', 'block');";

	return (n);
	}

function sendtext(class, options, n, text, keepopen,   i, cmd, line) {
#	printf ("<DIV CLASS=\"%s\">\n", class);
	cmd = wikiparser options;
#printf (">> cmd= %s, filename= %s\n", cmd, filename) >>STDERR;

	printf ("") |& cmd;
	printf ("<CONTENT-START>\n") |& cmd;

	for (i=1; i <= n; i++)
		printf ("%s\n", text[i]) |& cmd;

	close (cmd, "to");
	while (cmd |& getline line > 0)
		printf ("%s\n", line);

	close (cmd);
	if (0  &&  keepopen == 0) {
		printf ("</DIV>\n");
		printf ("<P>\n");
		}

	return (0);
	}

function sendfile(filename, options,    n, text) {
	n = readfile(filename, 0, text, 1, 0);
	sendtext("content", options, n, text, 0);
	return (0);
	}



#ifdef WITH_SUBDIRS

function getdirtree(basedir,   cmd, depth, dir, dirlen, line, list) {
	if (_dirtree != "")
		return (_dirtree);

	dir = basedir;

	basedir = basedir "/";
	dirlen = length(basedir);
	depth = 0;

	cmd = sprintf ("find '%s' -type d -print", basedir);

	list = ".";
	while (cmd |& getline line > 0) {
		if (line ~ /(^\.|\/\.| )/)
			continue;

		dir = substr(line, dirlen + 1);
		if (dir == "")
			continue;

		list = list " " dir;
		if (dir != ".") {
			gsub(/[^\/]/, "", dir);
			if (length(dir) > depth)
				depth = length(dir);
			}
		}

	close (cmd);
	_dirtree = depth " " list;
	return (_dirtree);
	}

function getsubdirs(dir,   cmd, line, list, n, x) {
	cmd = sprintf ("/bin/ls -l %s", quote(dir));
	while (cmd | getline line > 0) {
		n = split(line, x, " ");
		if (substr(x[1], 1, 1) != "d")
			continue;

		list = list " " x[n];
		}

	close (cmd);
	return (substr(list, 2));
	}

function getfilelist(basedir,   cmd, dirlen, file, line, list) {
	sub(/\/\.$/, "", basedir);
	basedir = basedir "/";
	dirlen = length(basedir);

	cmd = sprintf ("find '%s' -type f -maxdepth 1 -print | sort", basedir);
	while (cmd | getline line > 0) {
		if (line ~ /(^\.|\/\.| )/)
			continue;

		file = substr(line, dirlen + 1);
		list = list " " file;
		}

	close (cmd);
	return (substr(list, 2));
	}


function makedir(dir, name,   cmd) {
	subdir = dir "/" rewritename(name);
	cmd = sprintf ("/bin/mkdir %s", quote(wikipath "/" subdir));
	system(cmd);

	sendredirect(BASEADR name "?op=dirinfo");
	return (0);
	}

function removedir(dir, subdir,   cmd) {
	subdir = dir "/" rewritename(subdir);
	cmd = sprintf ("/bin/rmdir %s", quote(wikipath "/" subdir));
	system(cmd);

	sendredirect(BASEADR "?op=dirinfo");
	return (0);
	}

function renamedir(src, dest,   cmd, olddir, newdir) {
	olddir = wikipath "/" rewritename(src);
	newdir = wikipath "/" rewritename(dest);
	cmd = sprintf ("/bin/mv %s %s", quote(olddir), quote(newdir));
	system(cmd);

#	sendredirect(SCRIPTBASE dest "?op=dirinfo");
	sendredirect(BASEADR dest "?op=dirinfo");
	return (0);
	}

function printdirinfo(dir, subdir,   i, n, x, p, name) {
	printf ("<P>\n");
	printf ("Directory Info: <B>%s</B>\n", subdir);
	printf ("<P>\n");

	printf ("<U>Create New File:</U><P>\n");
	printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"new\"></INPUT>\n");
	printf ("  <DL>\n");
	printf ("  <DT>Enter filename for the file<DD>\n");
	printf ("  <INPUT TYPE=\"TEXT\" NAME=\"name\" VALUE=\"\" ACCESSKEY=\"N\" SIZE=\"30\">\n");
	printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Create\"></INPUT><BR>\n");
	printf ("  </DL>\n");
	printf ("</FORM>\n");

#ifdef WITH_NUMBERED
	printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
	printf ("  <DL>\n");
	printf ("  <DT>or the file's title to get a new numbered `-%s' file:<DD>\n", SERIALWIKINAME);
	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"new-numbered\">\n");
	printf ("  <INPUT TYPE=\"TEXT\" NAME=\"title\" VALUE=\"\" ACCESSKEY=\"U\" SIZE=\"30\">\n");
	printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Create\">\n");
	printf ("  </DL>\n");
	printf ("</FORM>\n");
#endif

	if (subdir != "") {
		name = subdir;
		sub(/^.*\//, "", name);

		printf ("<HR SIZE=\"1\">\n");
		printf ("<P>\n");
		printf ("<U>Rename or move directory:</U>\n");
		printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
		printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"renamedir\">\n");
		printf ("New name: <INPUT TYPE=\"TEXT\" NAME=\"name\" VALUE=\"\">\n");
		printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Rename\">\n");
		printf ("</FORM>\n");

		printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
		printf ("Select destination:\n");
		printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"movedir\">\n");
		printf ("<SELECT NAME=\"dest\">\n");
		printf ("  <OPTION VALUE=\"\">-- Select destination --</OPTION>\n");
		printf ("  <OPTION VALUE=\".\">%s</OPTION>\n", dirname[1]);

		n = split(p = getdirtree(wikipath "/" kbsubdir), x, " ");
		p = "^" subdir;
		for (i=3; i<=n; i++) {
			x[i] = x[i] "/";
			if (x[i] == subdir  ||  x[i] ~ p)
				continue;

			printf ("  <OPTION VALUE=\"%s\">%s</OPTION>\n", x[i], x[i]);
			}

		printf ("</SELECT>\n");
		printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Move\">\n");
		printf ("</FORM>\n");
		
		}
	printf ("<HR SIZE=\"1\">\n");
	printf ("<P>\n");
	printf ("<U>Create new directory:</U>\n");
	printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
	printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"mkdir\">\n");
	printf ("Directory name: <INPUT TYPE=\"TEXT\" NAME=\"name\" VALUE=\"\">\n");
	printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Create\">\n");
	printf ("</FORM>\n");

	printf ("<HR SIZE=\"1\">\n");
	printf ("<P>\n");
	printf ("<U>Remove subdirectory:</U>\n");
	printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
	printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"rmdir\">\n");
	printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"dir\" VALUE=\"%s\">\n", dir);
	printf ("Directory:\n");
	printf ("<SELECT NAME=\"subdir\"><BR>\n");
	printf ("  <OPTION VALUE=\"\">-- Select directory --</OPTION>\n");

	n = split(getsubdirs(dir), x, " ");
	for (i=1; i<=n; i++)
		printf ("  <OPTION VALUE=\"%s\">%s</OPTION>\n", x[i], x[i]);

	printf ("</SELECT>\n");
	printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Remove\">\n");
	printf ("</FORM>\n");

	return (0);
	}


#endif


function printselection(name, list, val, class,   i, k, n, v, x) {
	printf ("<SELECT NAME=\"%s\" CLASS=\"%s\">\n", name, class);
	n = split(list, x, "|");
	for (i=1; i <= n; i++) {
		k = index(x[i], ":");
		v = substr(x[i], 1, k-1);
		printf ("  <OPTION VALUE=\"%s\"%s>%s\n", v,
				v == val? " SELECTED": "", substr(x[i], k+1));
		}

	printf ("</SELECT>\n");
	return (0);
	}

function printsearchform(path, static, div,   base, targetseq, p, q) {
	path = getdirname(path) "/";
	if (option["w"] == ""  &&  option["s"] == "") {
		targetseq = "";
		base = makebaseurl(option, "", "");
		}
	else {
		targetseq = " TARGET=\"_parent\"";
		p = option["s"];  option["s"] = "";
		base = makebaseurl(option, "w", "");
		option["s"] = p;
		q = makebaseurl(option, "w", "");
		if (path != "")
			q = q path;
		}

#printf ("w= %s, targetseq= %s, base= %s\n", option["w"], targetseq, base) >>STDERR;
	printf ("<DIV ID=\"searchform\" CLASS=\"%s\" X-DISPLAY=\"%s\" ALIGN=\"LEFT\">\n",
			div == ""? "searchform": div, (static == 0? "none": "block"));

	printf ("<A NAME=\"search\"></A>\n");
	printf ("<U>Search: [%s]</U><P>\n", path);

	printf ("<FORM ACTION=\"%s%s\" NAME=\"search\" METHOD=\"GET\"%s onSubmit=\"settarget('%s');\">\n",
			base, path != "/"? path: "", targetseq, q);
	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"search\">\n");
	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"ph\" VALUE=\"1\">\n");

	printf ("  <TABLE CLASS=\"sc\" BORDER=\"0\">\n");

	if (userelsearch != 0) {
		printf ("    <TR>\n");
		printf ("      <TD CLASS=\"search\">Expression:</TD>\n");
		printf ("      <TD CLASS=\"search\">&#160; <INPUT TYPE=\"TEXT\" NAME=\"expr\" VALUE=\"%s\" ACCESSKEY=\"E\"></TD>\n", escape(ENVIRON["CGI_EXPR"], 1));
		printf ("    </TR>\n");
		}

	if (useregexpsearch != 0) {
		printf ("    <TR VALIGN=\"TOP\">\n");
		printf ("      <TD CLASS=\"search\">Pattern:</TD>\n");
		printf ("      <TD CLASS=\"search\">\n");
		printf ("        &#160; <INPUT TYPE=\"TEXT\" NAME=\"pattern\" VALUE=\"%s\" ACCESSKEY=\"P\">\n", escape(ENVIRON["CGI_PATTERN"], 1));
		printf ("        <BR>\n");
		printf ("        &#160; <INPUT TYPE=\"CHECKBOX\" NAME=\"case\" VALUE=\"yes\"%s>%s<BR>\n",
					(case == "yes")? " CHECKED": "", "match case");
		printf ("      </TD>\n");
		printf ("    </TR>\n");
		}

	printf ("<TR>\n");
	printf ("<TD CLASS=\"search\" COLSPAN=\"2\">\n");
	printf ("  <INPUT TYPE=\"CHECKBOX\" NAME=\"tv\" VALUE=\"yes\"%s> Display results in tree\n",
			(ENVIRON["CGI_TV"] != "")? " CHECKED": "");
	printf ("</TD>\n");
	printf ("</TR>\n");

	printf ("<TR>\n");
	printf ("<TD CLASS=\"search\" COLSPAN=\"2\">\n");
	printf ("  <INPUT TYPE=\"CHECKBOX\" NAME=\"cs\" VALUE=\"yes\"%s> Search whole server\n",
			(ENVIRON["CGI_CS"] != "")? " CHECKED": "");
	printf ("</TD>\n");
	printf ("</TR>\n");

	printf ("<TD CLASS=\"search\" COLSPAN=\"2\">\n");
		printf ("<NOBR><INPUT TYPE=\"CHECKBOX\" NAME=\"nw\" VALUE=\"new\"> Open new window</NOBR>\n");
	printf ("</TD>\n");

	if (isPDA == 0  &&  isPlain == 0) {
		printf ("<TR>\n");
		printf ("<TD CLASS=\"search\" ALIGN=\"MIDDLE\" COLSPAN=\"2\">\n");

		p = "-:-- framing: ";
		q = "";
		if (option["s"] != "") {
			p = p "|sw:this side|os:other side";
			q = "sw";
			}

		if (option["m"] != "sidebar")
			p = p "|rl:right/left|lr:left/right";

		p = p "|ud:up/down|du:down/up|new:new window";

		printselection("o", p, getval("CGI_O", option["o"], q), "search");
		printf ("&nbsp; &nbsp;\n");
		printselection("b", "-:-- results: |1.80:short|3.160:medium|5.300:long",
				getval("CGI_B", option["b"], "3.160"));
		printf ("</TD>\n");
		printf ("</TR>\n");
		}

	printf ("  </TABLE>\n");
	printf ("<P>\n");
#	if (side != "") {
#		printf ("  <INPUT TYPE=\"CHECKBOX\" NAME=\"target\" VALUE=\"%s\"%s>open files in %s frame<BR>\n",
#			tolower(otherside),
#			(target != "")? " CHECKED": "",
#			tolower(otherside));
#		}

	printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Search\"><BR>\n");
	printf ("</FORM>\n");
	printf ("</DIV>\n");

	return (0);
	}
 
function __getlinks(filename, list,   cmd, n, x, dir) {
	if (match(filename, /^.*\//) > 0)
		dir = "/" substr(filename, 1, RLENGTH-1);

	cmd = sprintf ("%s/wiki-parser -L <'%s/%s'", bindir, wikipath, filename);
	while (cmd | getline > 0) {
		split($0, x, "\t");
		if (x[1] !~ /[:?]/)
			$0 = wikipath dir "/" rewritename("/" x[1]);

		list[++n] = $0;
		}

	close (cmd);
	return (n);
	}

function __regexpsearch(base, title, fn, searchcmd, targetseq,   k, m) {
#printf ("grep= %s, cmd= %s\n", fn, searchcmd) >>STDERR;

	printf ("%s/%s\n", wikipath, fn) |& searchcmd;
	fflush(searchcmd);

	m = 0;
	while (searchcmd |& getline > 0) {
		if ($1 == ".")
			break;

		if (m == 0) {
			if (title != "")
				printf ("<P><LI>%s<BR>\n", title);

			printf ("<UL>\n");
			}

		k = eatone(1);
		printf ("<LI><A HREF=\"%s%s?line=%s#L%d\"%s>%s</A><BR>\n",
			base, fn, k-1, k-1, targetseq, $0);

		m++;
		}

	if (m > 0) {
		printf ("</UL>\n");
		printf ("<P>\n");
		}

	return (m);
	}

function __printresulttree(baseadr, n, list, targetseq, searchcmd,   subdir, i, k, m, file, set, p, q, x) {

	# Convert the result list into a associative array.  Fields
	# in the records are fullpath, filetext, rel-score and mime-type.

	m = length(wikipath);
	for (i=1; i<=n; i++) {
		split(list[i], x, "\t");
		p = substr(x[1], m+1);		# get relpath
		set[p] = list[i];
		}


	# Create a sorted list of files in the wikispace that
	# are also in the result set.  Take a directories into the
	# resulting list.  Use each matching filename to compute
	# the directory level and sortkey.

	# file[i] = k "\t" file[i] "\t" p "\t" m "\t" mime;

	cmd = sprintf ("/bin/ls -lR %s", quote(wikipath));
#printf ("cmd= %s\n", cmd) >>STDERR;
	while (cmd | getline > 0) {
		if ($1 ~ /:$/) {
			subdir = substr($1, length(wikipath) + 1, length($1) - length(wikipath) - 1) "/";
			m = split(subdir, x, "/");
			file[++n] = subdir "\002\t" subdir "\t" x[m-1] "/\t" m - 1 "\tdir";
			}
		else if (substr($1, 1, 1) != "-")
			continue;
		else if ((p = subdir $NF) in set) {
			split(set[p], x, "\t");
			file[++n] = subdir "/\002" $NF "\t" subdir $NF "\t" x[2] "\t" m "\t" x[4] "\t" x[3];
			}

#printf ("file= %s\n", file[n]) >>STDERR;
		}

	close (cmd);


	# Sort the array.

	IGNORECASE = 1;
	asort(file);

	# Create the HTML tree.

	k = 0;
	for (i=2; i<=n; i++) {
		if (file[i] == "")
			continue;

#printf ("file= %s\n", file[i]) >>STDERR;
		split(file[i], x, "\t");
		if ((m = x[4]) > k)
			printf ("<UL STYLE=\"line-height: 120%; margin-left: 1em; padding-left: .5em; x-list-style-type: none\">\n");
		else if (m < k) {
			while (k-- > m)
				printf ("</UL>\n");
			}

		printf ("<LI>");
		if (x[5] == "dir")
			printf ("<A HREF=\"%s\"%s>%s</A>\n", baseadr substr(x[2], 2), targetseq, x[3]);
		else {
			if (x[3] != "")
				printf ("%s<BR>\n", x[3]);

			p = x[2];
			sub(/^.*\//, "", p);
			printf ("<A HREF=\"%s\"%s>%s</A>\n", baseadr substr(x[2], 2), targetseq, p);
			if (x[6] != "") {
				split(x[6], y, " ");
				p = sprintf ("%s%c (%s)", y[4], "%", y[5]);
				p = (y[5] == "A"  ||  y[5] == "B")? "<B>" p "</B>": p;
				printf ("- %s, size= %s, score= %s\n", p, y[1], y[2]);
				}

			printf ("- %s\n", x[5]);

			if (searchcmd != "") {
				__regexpsearch(base, "", substr(x[2], 2), searchcmd, targetseq);
				}

			printf ("<P>\n");
			}

		k = m;
		}

	while (k-- > 0)
		printf ("</UL>\n");

	if (searchcmd != "")
		close (searchcmd);

	return (0);
	}

function __printlist(base, expr, n, list, targetseq, searchcmd, class,
					i, m, p, x, y) {
#	printf ("%s: %s<P>\n", word["resultsfor"], expr);
	printf ("<OL CLASS=\"%s\">\n", class == ""? "search": class);

	# fullpath -- filetext -- rel-score -- mime-type
	m = length(wikipath) + 2;
	for (i=1; i<=n; i++) {
		if (list[i] == "")
			continue;

		split(list[i], x, "\t");

		printf ("<LI>");
		if (x[1] ~ /[:?]/) {
			printf ("<A HREF=\"%s\"%s>%s</A>\n",
					x[1], (option["o"] != "")? " TARGET=\"_blank\"": targetseq,
					x[2]);
			}
		else {
			path = substr(x[1], m);
			if (x[2] != ""  &&  x[2] != "N/A")
				printf ("%s<BR>\n", x[2]);

			printf ("<A HREF=\"%s%s\"%s>%s</A>\n",
					base, path, targetseq, path);

			if (x[3] != "") {
				split(x[3], y, " ");
				p = sprintf ("%s%c (%s)", y[4], "%", y[5]);
				p = (y[5] == "A"  ||  y[5] == "B")? "<B>" p "</B>": p;
				printf ("- %s, size= %s, score= %s\n", p, y[1], y[2]);
				}

			printf ("- %s\n", x[4]);

			if (searchcmd != "")
				__regexpsearch(base, "", path, searchcmd, targetseq);
			}

		printf ("<P>\n");
		}

	printf ("</OL>\n");
	}

function dosearch(path, fn,   base, cmd, dir, type, expr, i, n, m, list, x,
				o, p, q, lw, searchcmd, targetseq) {
	base = BASEADR;

	sendheader(200, "text/html", 1);


	p = ENVIRON["CGI_PH"]+0;
	rmvar("ph");
	lw = (p == 1)? 30: 40;
	if (p >= 1  &&  ((o = ENVIRON["CGI_O"]) == "rl"  ||  o  == "lr"  ||  o == "du"  ||  o == "ud")) {
		rmvar("q1");
		printf ("<HTML>\n");
		printf ("<HEAD>\n");
		printf ("  <TITLE>%s - Search Results</TITLE>\n", wikiname);

		option["s"] = "";
		option["o"] = ENVIRON["CGI_O"];
		option["b"] = ENVIRON["CGI_B"];
		q = makebaseurl(option, "w", "list") path "?" ENVIRON["QUERY_STRING"];
		p = makebaseurl(option, "w", "text") getval("CGI_T", "", fn "?op=blank");

		if (o == "lr"  ||  o == "ud") {
			if (o == "lr")
				printf ("<FRAMESET COLS=\"%d%%,*\" FRAMEBORDER=\"1\" FRAMESPACING=\"2\">\n", lw);
			else
				printf ("<FRAMESET ROWS=\"%d%%,*\" FRAMEBORDER=\"1\" FRAMESPACING=\"2\">\n", searchheight);

			printf ("  <FRAME NAME=\"%s\" SRC=\"%s#results\">\n", frameside "list", q);
			printf ("  <FRAME NAME=\"%s\" SRC=\"%s\">\n", frameside "text", p);
			}
		else {
			if (o == "rl")
				printf ("<FRAMESET COLS=\"*,%d%%\" FRAMEBORDER=\"1\" FRAMESPACING=\"2\">\n", 30);
			else
				printf ("<FRAMESET ROWS=\"*,%d%%\" FRAMEBORDER=\"1\" FRAMESPACING=\"2\">\n", searchheight);

			printf ("  <FRAME NAME=\"%s\" SRC=\"%s\">\n", frameside "text", p);
			printf ("  <FRAME NAME=\"%s\" SRC=\"%s#results\">\n", frameside "list", q);
			}

		printf ("</FRAMESET>\n");
		printf ("</HEAD>\n");
		printf ("<BODY>\n");
		endhtml();

		exit(0);
		}

	starthtml("Search Pattern", "", 0, "");
	navigator(BASEADR path, "", 1);
	printsearchform(path, 1, "");


	if (wikimode == "pda"  ||  wikimode == "plain")
		targetseq = "";
	else if (ENVIRON["CGI_O"] == "os") {
		targetseq = sprintf(" TARGET=\"%s\"", otherside);
		option["o"] = "os";
		}
	else if (ENVIRON["CGI_O"] == "sw")
		option["o"] = "sw";
	else if (option["w"] == "list") {
		targetseq = sprintf(" TARGET=\"%s\"", "text");
		option["w"] = "text";
		}
	else if (ENVIRON["CGI_O"] == "new") {
		targetseq = " TARGET=\"_blank\"";
		option["w"] = option["m"] = "";
		}

	if (targetseq == ""  &&  option["m"] == "sidebar"  &&  isFirefox != 0)
		targetseq = " TARGET=\"_self\"";
		
	base = makebaseurl(option, "", "");
#printf ("targetseq= %s, base= %s\n", targetseq, base) >>STDERR;


	if ((o = option["o"]) == "du"  ||  o == "ud")
		printf ("<A NAME=\"results\"></A>\n");

	path = getdirname(path);
	expr = ENVIRON["CGI_EXPR"];
	type = ENVIRON["CGI_TYPE"];
	pattern = ENVIRON["CGI_PATTERN"];

	if (match(expr, /^(==?)(.*)$/, x) > 0) {
		expr = x[2];
		if (x[1] == "=")
			type = "n";
		else if (x[1] == "==")
			type = "l";

		if (type == "n"  &&  pattern != "")
			type = "l";
		}

	dir = wikipath;
	if (ENVIRON["CGI_CS"] != "yes")
		dir = wikipath "/" path;

#printf ("wikipath= %s, path= %s, dir= %s\n", wikipath, path, dir) >>STDERR;
	if (expr != "") {
		if (pattern != "") {
			searchcmd = sprintf ("%s -%s", parser, ENVIRON["CGI_CASE"] == "yes"? "r": "s");
			printf ("%s\n", pattern) |& searchcmd;
			}

		if (type == "n") {
			if ((p = rewritename("/" expr)) != "") {
#printf ("p= %s\n", p) >>STDERR;
				printf ("<DIV CLASS=\"browselist\">\n");
				q = targetseq;
				gsub(/(^[^"]*"|"[^"]*$)/, "", q);
				q = sprintf (" -T '%s'", q);
#				sendfile(wikipath "/" p, q);
				sendfile(p, q);
				}

			return (0);
			}
		else if (type == "l") {
			if ((p = rewritename("/" expr)) != "") {
#printf ("p= %s\n", p) >>STDERR;
				n = __getlinks(p, list);
#				__gettext(n, list);
#				__printlist(base, expr par, n, list, targetseq, searchcmd, "");
				}
			}
		else {
			cmd = sprintf ("find '%s' -type f ! -name '.*' | %s/wiki-relsearch -e %s",
					dir, bindir, quote(expr));
			while (cmd | getline > 0) {
				if (match($0, /^([^\t]+)\t(.*)$/, x) > 0)
					list[++n] = $0;
				}

			close (cmd);
			}

		__gettext(n, list, "");
		if (ENVIRON["CGI_TV"] == "yes")
			__printresulttree(base, n, list, targetseq, searchcmd);
		else
			__printlist(base, expr, n, list, targetseq, searchcmd, "");
		}
	else if (pattern != "") {
		n = getdirfilelist(path, list);
		m = length(wikipath) + 1;

		searchcmd = sprintf ("%s -%s", parser, ENVIRON["CGI_CASE"] == "yes"? "r": "s");
		printf ("%s\n", pattern) |& searchcmd;

		printf ("<OL>\n");
		for (i=1; i<=n; i++) {
#printf ("file= %s\n", list[i]) >>STDERR;
			split(list[i], x, "\t");
			if (x[4] != "text/yawk")
				continue;

			__regexpsearch(base, x[2], substr(x[1], m+1), searchcmd, targetseq);
			}

	#	__regexpsearch("", path, searchcmd, targetseq);
		printf ("</OL>\n");
#		close (searchcmd);
		}

	printf ("</DIV>\n");
	navigator(BASEADR path, "", 0);
	endhtml();

	if (searchcmd != "")
		close (searchcmd);

	return (0);
	}


function openform(base, path,   tg, file) {
	base = BASEADR;
	tg = "";
	if ((p = option["w"]) != "") {
		tg = " TARGET=\"_parent\"";
		base = makebaseurl(option, "w", "");
		}
	else if (applyfirefoxpatch != 0)
		tg = " TARGET=\"_self\"";

	file = getbasename(path);
	path = getdirname(path) "/";

	printf ("<DIV ID=\"openform\" CLASS=\"%s\">\n", "openform");
	printf ("<U>Open: [%s]</U><P>\n", path);
	printf ("<FORM NAME=\"open\" STYLE=\"display: inline;\" ACTION=\"%s\"%s onsubmit=\"settarget();\">\n",
			base, targetseq);

	if (p != "")
		base = makebaseurl(option, "w", p);

	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"open\">\n");
	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"ph\" VALUE=\"1\">\n");
	printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"pa\" VALUE=\"%s\">\n", path);

	printf ("Filename: ");
	printf ("  <INPUT TYPE=\"TEXT\" NAME=\"s\" SIZE=\"20\" VALUE=\"%s\" ACCESSKEY=\"%s\">\n",
			file, ak["open"]);
	printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"q1\" VALUE=\"%s\">\n", word["open"]);
	printf ("<BR>\n");

	printf ("<NOBR><INPUT TYPE=\"CHECKBOX\" NAME=\"edit\" VALUE=\"yes\"> Open in editor</NOBR>\n");
	printf ("<BR>\n");

	printf ("Options:\n");
	printselection("fs", "-:-- font: |8:8|9:9|10:10|11:11|12:12|13:13",
			getval("CGI_FS", option["fs"], ""));
	printf (" -- ");
	printselection("m", "-:-- mode: |pda:PDA|sidebar:sidebar|plain:plain|offline:offline|crawler:crawler", getval("CGI_M", option["m"], ""));
	printf (" -- ");

	printf ("<NOBR><INPUT TYPE=\"CHECKBOX\" NAME=\"abs\" VALUE=\"hide\"%s> Hide abstracts</NOBR>\n",
			option["abs"] == "hide"? " CHECKED": "");


	printf ("</FORM>\n");
	printf ("</DIV>\n");

	return (0);
	}

function doopen(dir,   base, path, qs) {
	path = ENVIRON["CGI_S"];
	if (substr(file, 1, 1) != "/")
		path = getval("CGI_PA", "/", "") "/" path;

	option["fs"] = getval("CGI_FS", "", "");
	option["m"] = getval("CGI_M", "", "");
	option["abs"] = getval("CGI_ABS", "", "");
	base = makebaseurl(option, "", "");

	path = rewritename("/" path);
	if (ENVIRON["CGI_EDIT"] != "")
		qs = "?op=edit";

	sendredirect(base path qs, "", "");
	return (0);
	}


function __getfiles(n, list, text, relp,   i, m, lines, done) {
	lines = 0;
	for (i=1; i <= n; i++) {
		if (list[i] in done)
			continue;

#printf (">> readfile: %s\n", list[i]) >>STDERR;
		m = lines;
		lines = readfile(list[i], lines, text, relp, 1);
		if ("noprint" in fileconf)
			lines = m;

#printf (">> m= %d, lines= %s, readfile: %s\n", m, lines, list[i]) >>STDERR;
		done[list[i]] = 1;
		}

	return (lines);
	}

function docompound(class, options, filename,   i, k, n, list, text, p) {
	filename = rewritename("/" filename);
#printf (">> filename= %s\n", filename) >>STDERR;
	n = __getlinks(filename, list);
	k = 0;
	for (i=1; i <= n; i++) {
		if (list[i] !~ /[:?]/) {
			list[++k] = list[i];
#printf (">> k= %d, list[k]= %s\n", k, list[k]) >>STDERR;
			}
		}

	n = k;
	lines = __getfiles(n, list, text, 0);

	printf ("<DIV CLASS=\"%s\">\n", "compound");
	sendtext(class, options, lines, text, 0);

	return (0);
	}


function setlist(arr, list,   i, k, n, x) {
	n = split(list, x, /[ ,]+/);
	k = 1;
	for (i=1; i<=NF; i++) {
		if (x[i] == "-") {
			delete arr;
			k = 0;
			}
		else {
			arr[x[i]] = 1;
			k = 1;
			}
		}

	return (k);
	}


function getdirfilelist(path, list, sb,   cmd, i, m, n, p, x) {
	if (path !~ /\/$/)
		path = path "/";

	cmd = sprintf ("/bin/ls -ld %s*", quote(wikipath "/" path));
#printf ("path= %s, cmd= %s\n", path, cmd) >>STDERR;
	while (cmd | getline line) {
		m = split(line, x, /[ ]+/);
		p = x[m];
#printf ("p= %s\n", p) >>STDERR;
		__computestat(p, sb, line);
		if (sb[p, "type"] == "file")
			list[++n] = p;
		else if (sb[p, "type"] == "dir")
			list[++n] = p "/";
		}

	close (cmd);
	__gettext(n, list, "3,120");

	# sort array by file title (col 2).
	for (i=1; i<=n; i++) {
		split(list[i], x, "\t");
		list[i] = x[2] "\t" list[i];
		}

	IGNORECASE = 1;
	n = asort(list);
	for (i=1; i<=n; i++) {
		sub(/^[^\t]*\t/, "", list[i]);
#printf ("i= %d, line= %s\n", i, list[i]) >>STDERR;
		}

	# returns: (fullpath -- title text -- .unused. -- mimetype)
	return (n);
}

function dobrowser(path,   i, n, m, list, p, x, cmd, stat, targetseq) {
	n = getdirfilelist(path, list, stat);

	sendheader(200, "text/html", 1);
	starthtml(path, cssadr, 0);
	readonly = 1;
	navigator(BASEADR, "", 1);

	if (applyfirefoxpatch != 0)
		targetseq = " TARGET=\"_self\"";

	printf ("<DIV CLASS=\"dir\">\n");
	printf ("&#160;<P>\n");
	printf ("<TABLE CLASS=\"browser\" CELLSPACING=\"0\" WIDTH=\"90%\">\n");

	printf ("<TR BGCOLOR=\"#C0C0C0\"><TD><B>%s</B></TD></TR>\n", mksubdirlocation("", wikisubdir, "?op=browse", "", "?op=browse", ""));
	for (i=1; i<=n; i++) {
		split(list[i], x, "\t");
		p = getbasename(x[1]);

		printf ("<TR CLASS=\"dir\"%s>\n", (i % 2) == 0? " BGCOLOR=\"#F0F0F0\"": "");
		printf ("  <TD VALIGN=\"TOP\" STYLE=\"overflow: hidden;\"><NOBR>\n");

		if (x[4] == "dir")
			printf ("    <A HREF=\"%s?op=browse\"%s CLASS=\"browser\">%s</A>\n", x[2], targetseq, x[2]);
		else
			printf ("    <A HREF=\"%s\"%s CLASS=\"browser\">%s</A>\n", p, targetseq, x[2]);

		printf ("  </NOBR></TD>\n", x[2]);
		printf ("</TR>\n");
		}

	printf ("</TABLE>\n");
	printf ("</DIV>\n");

	navigator(BASEADR, "", 0);
	endhtml();
	}

function dolist(path,   i, n, m, list, p, x, cmd, stat, targetseq) {
	n = getdirfilelist(path, list, stat);
	asort(list);

	sendheader(200, "text/html", 1);
	starthtml(path, cssadr, 0);
	readonly = 1;
	navigator(BASEADR, "", 0);

	if (applyfirefoxpatch != 0)
		targetseq = " TARGET=\"_self\"";

	printf ("<DIV CLASS=\"dir\">\n");
	printf ("&#160;<P>\n");
	printf ("<TABLE CLASS=\"dir\" CELLSPACING=\"0\">\n");

	printf ("<TR><TD COLSPAN=\"4\">%s</TD></TR>\n", subdirlocation);
	printf ("<TR CLASS=\"dirtitle\" BGCOLOR=\"#F0F0F0\">\n");
	printf ("  <TD>Title</TD><TD>Filename</TD><TD>Size</TD><TD>Date</TD>\n");
	printf ("</TR>\n");

	for (i=1; i<=n; i++) {
		split(list[i], x, "\t");
		p = getbasename(x[1]);

		printf ("<TR CLASS=\"dir\"%s>\n", (i % 2) == 0? " BGCOLOR=\"#F0F0F0\"": "");
		printf ("  <TD VALIGN=\"TOP\">%s</TD>\n", x[2]);
		printf ("  <TD VALIGN=\"TOP\"><NOBR><A HREF=\"%s\"%s>%s</A></NOBR></TD>\n", p, targetseq, p);
		printf ("  <TD VALIGN=\"TOP\" ALIGN=\"RIGHT\">%s</TD>\n", stat[x[1], "size"]);
		printf ("  <TD VALIGN=\"TOP\"><NOBR>&#160;%s</NOBR></TD>\n",
				strftime("%Y-%m-%d %H:%M", stat[x[1], "mtime"]));
		printf ("</TR>\n");
		}

	printf ("</TABLE>\n");
	printf ("</DIV>\n");
	endhtml();
	}

function doindex(fn, type,    i, k, n, m, p, list, sbuf, havelink, x) {
	p = substr(fn, length(wikipath) + 2);
#printf ("fn= %s, p= %s\n", fn, p) >>STDERR;
	n = __getlinks(p, list);
	for (i=1; i<=n; i++) {
#printf ("list[i]= %s\n", list[i]) >>STDERR;
		havelink[list[i]] = 1;
		}

	p = getdirname(fn) "/";
	n = split(fileconf["isindex"], x, / +/);
	for (i=1; i<=n; i++) {
		if (x[i] == "")
			continue;

		havelink[p x[i]] = 1;
		}

	p = substr(getdirname(fn), length(wikipath) + 2);
	n = getdirfilelist(p, list, sbuf);
	for (i=1; i<=n; i++) {
		split(list[i], x, "\t");
#printf ("fn= %s\n", x[1]) >>STDERR;
		if (x[1] == fn  ||  x[1] in havelink)
			list[i] = "";
		}

	printf ("<%s class=\"filelist\">\n", type);
	for (i=1; i<=n; i++) {
		if (list[i] == "")
			continue;

		split(list[i], x, "\t");
		p = substr(x[1], length(wikipath) + 2);

		printf ("<LI>%s<BR>\n", x[2]);
		printf ("<A HREF=\"%s\">%s</A> - %s\n", BASEADR p, getbasename(p), x[4]);
		printf ("<P>\n");
		}

	printf ("</%s>\n", type);
#	for (i=1; i<=n; i++)	# prepare for __printlist()
#		sub(/\t[^\t]*$/, /\t\t&/, list[i]);
#
#	__printlist(BASEADR, "", n, list, "", "", "filelist");
	return (0);
	}


function __printtree(basedir, dir,   subdir, i, k, n, m, file, mime, p, q, x) {

	# Create a sorted list of all files in the wikispace.

	q = basedir;
	if (dir != "")
		q = q "/" dir;

	cmd = sprintf ("/bin/ls -lR %s", quote(q));
#printf ("cmd= %s\n", cmd) >>STDERR;
	while (cmd | getline > 0) {
		if ($1 ~ /:$/) {
			if ($1 ~ /\/\./)
				continue;

			subdir = substr($1, length(basedir) + 1, length($1) - length(basedir) - 1) "/";
			sub(/^\/+/, "/", subdir);
			file[++n] = subdir;
			continue;
			}
		else if (substr($1, 1, 1) != "-"  ||  $NF ~ /^\./)
			continue;
		else
			file[++n] = subdir $NF;
#printf ("file= %s\n", file[n]) >>STDERR;
		}

	close (cmd);
	n = asort(file);

	# Get the titles from all wikifiles.

	cmd = wikiparser " -bb 1,80";
#printf ("cmd= %s\n", cmd) >>STDERR;
	printf ("") |& cmd;
	for (i=1; i<=n; i++) {
		m = split(file[i], x, "/") - 1;	# Compute directory level

		if (file[i] ~ /\/$/) {
			p = x[m] "/";
			m -= 1;
			mime = "dir";
			}
		else if ((mime = getmimetype(x[m+1])) != "text/yawk")
			p = x[m+1];
		else {
			printf ("%s\n", basedir file[i]) |& cmd;
			fflush(cmd);
			cmd |& getline p;
			sub(/^.*\t/, "", p);
			}

#printf ("file= %s, p= %s, m= %d\n", file[i], p, m) >>STDERR;
		k = file[i];		# Compute sort key
		sub(/\/[^\/]*$/, "", k);
		k = k "/\002" (mime == "dir"? "": p);

		file[i] = k "\t" file[i] "\t" p "\t" m "\t" mime;
#printf ("file= %s\n", file[i]) >>STDERR;
		}

	close (cmd);
	IGNORECASE = 1;
	asort(file);

	# Create the HTML tree.

	q = split(q, x, "/");
	k = 0;
	for (i=2; i<=n; i++) {
		split(file[i], x, "\t");
		if ((m = x[4]) > k) {
			while (m > k++)
				printf ("<UL line-height: 120%; margin-left: 1em; padding-left: .5em; x-list-style-type: none\">\n");
			}
		else if (m < k) {
			printf ("<P>\n");
			while (k-- > m)
				printf ("</UL>\n");
			}

		printf ("<LI><A HREF=\"%s\"%s>%s</A>\n", baseadr substr(x[2], 2), targetseq, x[3]);
		k = m;
		}

	while (k-- > 0)
		printf ("</UL>\n");

	return (0);
	}

function __doinfopage(fn,   bn, url) {
	url = BASEADR path;
	bn = getbasename(fn);

	if (path == "")
		printf ("Location: %s<BR>\n", subdirlocation);
	else {
		printf ("<U>Fileinfo:</U> <A HREF=\"%s\">%s</A><P>\n", url, bn);

		printf ("Location: %s<BR>\n", subdirlocation);
		printf ("Size: %d bytes<BR>\n", sbuf[fn, "size"]);
		printf ("Last modified: %s<BR>\n", sbuf[fn, "lastmod"]);
		}

	if (path != "") {
		if (readonly == 0) {
			printf ("<A HREF=\"%s?op=edit\">Edit</A> file.\n", url);
			if (side != "") {
				printf ("Edit <A HREF=\"%s?op=edit&target=%s\">%s to %s</A>.\n", url, otherside,
						tolower(side), tolower(otherside));
				}
			}

		printf ("<P>\n");
		printf ("<A HREF=\"%s?\">Reload</A> file.\n", url);
		}


	if (side != "") {
		printf ("Load into <A HREF=\"%s\" TARGET=\"%s\">%s side</A>,",
			makebaseurl(option, "s", otherside) path,
			otherside, tolower(otherside));

		url = makebaseurl(option, "s", "") path;
		printf (" <A HREF=\"%s\" TARGET=\"_parent\">unsplit screen</A>.\n", url);
		}
	else if (isPDA == 0  &&  isPlain == 0  &&  option["m"] != "sidebar") {
		printf ("<A HREF=\"%s?op=frame\" TARGET=\"_parent\">Split screen</A>.\n", url);
		}


	if (path != "") {
		if (wikimode == "") {
			printf ("<BR>Bookmark as <A HREF=\"%s\" REL=\"sidebar\" TITLE=\"%s\">sidebar</A>.\n",
					makebaseurl(option, "m", "sidebar") path, wikiname);
			}

		printf ("<BR>Display <A HREF=\"%s?op=printer\" TARGET=\"_blank\">printer friendly</A>\n", url);
		if (csspath != "")
			printf ("or <A HREF=\"%s?op=save-as\" TARGET=\"_blank\">\"save-as\"</A>\n", url);
		printf ("version.\n", fileurl);

		printf ("<BR>Download or display <A HREF=\"%s?op=as-binary\" TARGET=\"_blank\">unformatted</A>.\n", url);
#		printf ("<BR>List <A HREF=\"%s?op=search&pattern=%s&case=yes\">wiki references</A> to this file.\n",
#				path, path);
		}

	printf ("<HR SIZE=\"1\">\n");
	openform("", path);

	printf ("<HR SIZE=\"1\">\n");
	printsearchform(path, 1, "");

	if (path != "") {
		if (readonly == 0) {
			printf ("<HR SIZE=\"1\">\n");
			printf ("<U>Rename or copy:</U><P>\n"); 
			printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", url);
			printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"rename\">\n");
			printf ("  <INPUT TYPE=\"TEXT\" NAME=\"target\" VALUE=\"%s\" ACCESSKEY=\"R\">\n", bn);
			printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Rename\"><BR>\n");
			printf ("  <INPUT TYPE=\"CHECKBOX\" NAME=\"copy\" VALUE=\"yes\"> create copy\n");
			printf ("</FORM>\n");
			printf ("Use the sequence \"---\" (three dashes) to delete the file.\n");
			}
		}


	if (readonly == 0) {
		if (uploadfiles == 1) {
			printf ("<HR SIZE=\"1\">\n");

			printf ("<U>File Upload:</U><P>\n"); 
			printf ("<FORM ACTION=\"%s00-FileUpload\" METHOD=\"POST\" ENCTYPE=\"multipart/form-data\">\n", BASEADR);
			printf ("  <INPUT TYPE=\"FILE\" NAME=\"file\" SIZE=\"30\" ACCESSKEY=\"U\"> &#160;\n");
			printf ("  <INPUT TYPE=\"SUBMIT\" VALUE=\"Upload\">\n");
			printf ("</FORM>\n");
			}
		}

	if (readonly == 0) {
		if (synccmd != ""  ||  exportcmd != "") {
			printf ("<HR SIZE=\"1\">\n");

			printf ("<U>Syncronize or export:</U><P>\n"); 
			printf ("<FORM ACTION=\"%s%s\" METHOD=\"POST\">\n", BASEADR, path);
			printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"sync\">\n");
			printf ("  <INPUT TYPE=\"PASSWORD\" NAME=\"passwd\" VALUE=\"\" ACCESSKEY=\"Y\">\n");

			if (synccmd != "")
				printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Sync\">\n");

			if (exportcmd != "")
				printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Export\"><BR>\n");

			printf ("</FORM>\n");
			}
		}

	navigator(BASEADR, "", 0);
	endhtml();
	}


function setword(key, str,   x) {
	if (match(str, /^([^&]*)&(.)(.*)$/, x) == 0) {
		ak[key] = "";
		}
	else if (index(str, ";") == 0){
		ak[key] = toupper(x[2]);
		str = x[1] x[2] x[3];
		}

	word[key] = str;
	return (0);
	}

function readconfig(dir, config, mode, req, included,   filename, n, p, line, lines) {
	if (config == "")
		config = "wiki";

	if (included != 0)
		filename = config;
	else {
		filename = sprintf ("%s/%s-wiki.conf", dir, config);

		wikiname = config;
#ifdef WITH_DOWNLOADABLE
		csspath = "";
#endif
		showsearch = 0;

#ifdef WITH_NUMBERED
		SERIALFILENAME = ".serial";
		SERIALWIKINAME = "WikiFile";
#endif

		wraptext = 0;

		useregexpsearch = 1;
		userelsearch = 1;
		showdirnav = 1;
		showtopnav = 1;
		uploadfiles = 1;

		fontfamily = "Verdana, Georgia, Helvetica, Arial, sans-serif";
		charset = "iso-8859-1";
		internalstyle = 1;
		listfontsize = "100%";
		personalmode = 0;
		readonly = 0;
		searchheight = 50;
		send404 = 0;
		wikiindex = "00-WikiIndex";
		usetablenav = 1;

		rellist = "home:pnx index:pnx contents:pnx search:pnx glossary:pnx help:pnx:_blank" \
				" first:pnx prev:pnx next:pnx last:pnx up:pnx" \
				" copyright:pnx author:pnx";

		n = split("top:TOP,index:&Index,list:&List,edit:&Edit,open:&Open,save:&Save" \
			  ",close:&Close,reset:&Reset,search:&Search" \
			  ",info:Inf&o,lock:Loc&k,unlock:Unloc&k,pnx:PNX,home:&Home" \
			  ",resultsfor:Search results for:" \
			  ",first:First,last:Last,prev:&Prev,next:&Next,up:Up" \
			  ",contents:Con&tents,index:Index,help:Help" \
			  ",copyright:Copyright,author:Author" \
			  ",navigator:Navigator,linker:Links,compound:Compound", x, ",");

		for (i=1; i<=n; i++) {
			split(x[i], y, ":");
			setword(y[1], y[2]);
			}
		}



	lines = skip = 0;
	while (getline <filename > 0) {
		$0 = noctrl($0);
		lines++;

		if ($0 == "")
			skip = 0;
		else if (skip != 0)
			continue;

		if ($1 ~ /^#/)
			continue;

		if (match($0, /^[ \t]*\[(.*)\](.*)$/, x) > 0) {
			$0 = skipws(x[2]);
			n = split(q = x[1], x, /[ ,]+/);
		
			skip = -1;
			c = 0;
			for (i=1; i <= n; i++) {
				if ((p = skipws(x[i])) == "")
					continue;

				if ((c = substr(p, 1, 1)) == "-"  ||  c == "+")
					p = substr(p, 2);
				else
					c = "+";

				if (substr(p, 2, 1) != "=")
					m = "m";
				else {
					m = substr(p, 1, 1);
					p = substr(p, 3);
					}

				c = (c == "+")? 0: 1;
				if (m == "r")
					skip = match(req, p, up) > 0? c: 1-c;
				else if (m == "m")
					skip = (p == "*"  ||  p == mode)? c: 1-c;

#printf ("x[i]= %s, c= %d, m= %s, p= %s, skip= %s\n", x[i], c, m, p, skip) >>STDERR;
				if (skip == 1)
					break;
				}

			if (skip < 0)
				skip = 1;		# skip = 1 - c;
#printf ("q= %s, skip= %s\n", q, skip) >>STDERR;
			continue;
			}


		p = eatone(1);
		gsub(/-/, "", p);
		if (p == "dir"  &&  included == 0)
			wikipath = $1;

		else if (p == "index")
			wikiindex = $2;

		else if (p == "name"  ||  p == "wikiname")
			wikiname = $0;
		else if (p == "personalmode") {
			if ($1 == "yes")
				personalmode = 1;
			else if ($1 == "closed")
				personalmode = 2;
			else if ($1 == "no")
				personalmode = 0;
			}

		else if (p == "rel") {
			p = eatone(1, 0);
			sub(/:$/, "", p);
			fileconf[p] = $0;
			}
		else if (p == "internal-style"  ||  p == "internal-styles")
			internalstyle = ($1 == "yes"? 1: 0);
		else if (p == "style")
			cstyle[++nstyle] = $0;

		else if (p == "printercss")
			printercss = $1;
		else if (p == "charset")
			charset = $1;
		else if (p == "csspath")
			csspath = $1;
		else if (p == "script")
			scriptfile[++scripts] = $1;
		else if (p == "handler"  ||  p == "jsc")
			bodyhandler[++jshc] = $0;
		else if (p == "textarea") {
			rows = $1;
			cols = $2;
			}
		else if (p == "css") {
			for (i=1; i<=NF; i++)
				stylesheet[++csscount] = $i;
			}
		else if (p == "showsearch") {
			if ($1 == "yes")
				showsearch = 1;
			else if ($1 == "input")
				showsearch = 2;
			else
				showsearch = 0;
			}
		else if (p == "usetablenav")
			usetablenav = ($1 == "yes")? 1: 0;

		else if (p == "listfontsize")
			listfontsize = $1;
		else if (p == "searchheight")
			searchheight = $1+0;

		else if (p == "header") {
			if ((p = $1) ~ /^(\.|-|no|off|none)$/)
				wikiheader = "";
			else
				wikiheader = $1;
			}
		else if (p == "footer") {
			if ((p = $1) ~ /^(\.|-|no|off|none)$/)
				wikifooter = "";
			else
				wikifooter = $1;
			}
		else if (p == "send404")
			send404 = $1 == "yes"? 1: 0;

		else if (p == "sync")
			synccmd = $0;
		else if (p == "export")
			exportcmd = $0;
		else if (p == "type") {
			type = eatone(1);
			line = skipws(noctrl($0));
			gsub(/[ ,]+/, " ", line);
			usertype[type] = " " line " ";
			}

		else if (p == "writeperm")
			havewriteperms = setlist(writeperm, $0);
		else if (p == "readonly")
			readonly = ($1 == "yes")? 1: 0;

		else if (p == "word") {
			p = eatone(1, 0);
			setword(p, $0);
			}

		else if (p == "include") {
			p = eatone(1, 0);
			if (substr(p, 1, 1) != "/")
				p = wikipath "/" p;

			readconfig(dir, p, mode, req, 1);
			}
		}

	close (filename);
	if (subconfig == 0) {
		if (lines == 0) {
			sendheader(200, "text/html", 0);
			starthtml("Wiki Error", "", 0, "");
			printf ("No such wiki: %s\n", config);
			endhtml();
			exit (0);
			}

		wikiconfigfile = filename;
		}

	if (wikiindex == "")
		wikiindex = "00-WikiIndex";

	if (listfontsize+0 == 0)
		listfontsize = "100%";

	if (searchheight < 10)
		searchheight = 10;

	return (0);
	}



function __insert_editform(file, n, text, target,   i) {
	targetadr = BASEADR path;
	targetopt = "";
	if (target != "") {
		targetopt = sprintf (" TARGET=\"%s\"", target);
		targetadr = makebaseurl(option, "s", otherside) path;
		}

#printf ("file= %s, n= %s\n", file, n) >>STDERR;
	datelastmod = sbuf[file, "mtime"];

	printf ("<FORM ACTION=\"%s\" METHOD=\"POST\"%s>\n", targetadr, targetopt);
	printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"%s\">\n", "store");

	printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"datelastmod\" VALUE=\"%s\">\n", datelastmod);
	printf ("<INPUT TYPE=\"HIDDEN\" NAME=\"target\" VALUE=\"%s\">\n", target);

	if (rows > 0  &&  cols > 0)
		printf ("<TEXTAREA NAME=\"text\" ROWS=\"%d\" COLS=\"%d\" CLASS=\"edit\" ACCESSKEY=\"E\">", rows, cols);
	else
		printf ("<TEXTAREA CLASS=\"edit\" NAME=\"text\" ACCESSKEY=\"E\">");

	for (i=1; i<=n; i++) {
		line = text[i];
#		if ((line = text[i]) ~ /^:[a-z]+:/)
#			continue;

		gsub(/&/, "\\&amp;", line);
		gsub(/>/, "\\&gt;", line);
		gsub(/</, "\\&lt;", line);
		printf ("%s\r\n", noctrl(line));
		}

	printf ("</TEXTAREA><P>\n");
	printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"preview\" VALUE=\"Preview\">\n");
	printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Save\">\n");

	if (target == "")
		printf ("<INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Close\">\n");

	if (datelastmod != ""  &&  targetopt == "") {
		printf (" &#160; <INPUT TYPE=\"CHECKBOX\" NAME=\"overwrite\" VALUE=\"yes\">\n");
		printf (" Overwrite changes");
		}

	printf ("</FORM>\n");
	return (0);
	}


function mkoptions(config, mode, side, other,   o) {
	if (config == ""  &&  mode == ""  &&  side == ""  &&  other != "")
		return ("");

	if (config == "")
		config = "wiki";

	o = "-" config;
	if (mode != ""  ||  side != ""  ||  other != "") {
		o = o configdelim mode;
		if (side != ""  ||  other != "") {
			o = o configdelim side;
			if (other != "")
				o = o configdelim other;
			}
		}

	if (o != "")
		o = o "/";

	return (o);
	}

function nextarg(par,   arg) {
	if (argi >= ARGC) {
		printf ("wiki.cgi: missing argument: %s\n", par) >"/dev/stderr";
		exit (1);
		}
		
	arg = ARGV[argi];
	ARGV[argi++] = "";

	return (arg);
	}

BEGIN {
	STDERR = "/dev/stderr";
	wikiparser = parser = sprintf ("%s/wiki-parser", bindir);
	respstatus = respcontent = "-";

	CamelCase = "^(([0-9]+|[a-z]+)-?)?[A-Z]+[a-z0-9]+([A-Z]+[a-z0-9]+)+$";
	BASEADR = ENVIRON["SCRIPT_NAME"];

	if (configdelim == "")
		configdelim = ",";


	sub(/\/+$/, "", BASEADR);
	BASEADR = BASEADR "/";

	selector = ENVIRON["PATH_INFO"];
	if (selector == "") {
		sendredirect(BASEADR, "empty or unset PATH_INFO");
		exit (0);
		}

	config = side = wikimode = options = "";
#	if (match(selector, /^\/-[^\/]*\//)) {
#		options = substr(selector, RSTART+2, RLENGTH-3);
#		selector = substr(selector, RSTART+RLENGTH-1);
#
#		n = split(options, x, configdelim);
#		config = x[1];
#		wikimode = tolower(x[2]);
#		side = tolower(x[3]);
#
#		options = "";
#		for (i=4; i<=n; i++) {
#			if (x[i] == "locked")
#				wikiopt["lockframe"] = "locked";
#			else if (x[i] == "raw")
#				wikiopt["displaymode"] = "raw";
#			else
#				continue;
#
#			options = options "," x[i];
#			}
#
#		if (wikiopt["lockframe"] == "")
#			wikiopt["lockframe"] = "unlocked";
#
#
#		if (selector == "") {
#			sendredirect(BASEADR, "empty selector");
#			exit (0);
#			}
#
#		if (side == "up"  ||  side == "down") {
#			wikimode = "sidebar";
#			}
#		else if (side != "") {
#			side = tolower(side);
#			otherside = "left";
#			if (tolower(side) == "left")
#				otherside = "right";
#			}
#
#		BASEADR = sprintf ("%s/%s", ENVIRON["SCRIPT_NAME"], mkoptions(config, wikimode, side));
#		}
#
	config = "wiki";
	if (match(selector, /^\/-([^\/]*)(\/?.*$)/, y) > 0) {
		options  = y[1];
		selector = y[2];

		n = split(options, x, /,+/);
		config = x[1];
		for (i=2; i<=n; i++) {
			if (match(x[i], /^[_a-zA-Z][_a-zA-Z0-9]*$/) > 0)
				option[x[i]] = "";
			else if (match(x[i], /^([_a-zA-Z][_a-zA-Z0-9]*)=(.*)$/, y) > 0)
				option[y[1]] = y[2];

			requestoptions++;
			}
			
#		BASEADR = sprintf ("%s/-%s%s/", ENVIRON["SCRIPT_NAME"], config, makeoptions(option, "", ""));
		BASEADR = makebaseurl(option, "", "");
## ???		frameside = option["s"];

		if ("m" in option) {
			wikimode = option["m"];
			if (wikimode == "crawler")
				requestoptions--;
			}

		if ("s" in option) {
			if ((side = option["s"]) == "left")
				otherside = "right";
			else
				otherside = "left";
			}
		}


#	SCRIPTBASE = BASEADR;

	if (ENVIRON["REQUEST_METHOD"] == "GET")
		getcgivars(ENVIRON["QUERY_STRING"]);
	else if (ENVIRON["REQUEST_METHOD"] == "POST") {
		if (ENVIRON["CONTENT_TYPE"] ~ /^multipart\/form-data/)
			;
		else {
			cmd = sprintf ("dd bs=1 count='%d' 2>/dev/null", ENVIRON["CONTENT_LENGTH"]);
			cmd | getline POSTDATA;
			close(cmd);
			getcgivars(POSTDATA);
			}
		}


	if (selector == ""  ||  (p = rewritename(selector)) == "\001") {
		sendredirect(BASEADR);
		exit (0);
		}

	readconfig(configdir, config, wikimode == ""? "normal": wikimode, "/" p);

	if (personalmode == 2) {
		if (ENVIRON["REMOTE_ADDR"] != "127.0.0.1")
			exit (0);
		}

	if (wikimode == "sidewiki"  ||  wikimode == "sidebar") {
		if (sidebarcss != "")
			cssadr = sidebarcss;
		}


	title = path = rewritename(selector);
	sub(/^.*\//, "", title);
	wikisubdir = getdirname(path) "/";
	if (wikisubdir == "/")
		wikisubdir = "";


	# Check if it's a webcrawler ...
	p = tolower(ENVIRON["HTTP_USER_AGENT"]);
	if (p ~ /msnbot|googlebot|yahoo! slurp|voilavot|furlbot|turnitinbot|mj12bot|exabot/)
		isCrawler = 1;
	else if (ENVIRON["CGI_OP"] == "crawler")
		isCrawler = 1;

	# ... and process it.
	if (isCrawler != 0  ||  wikimode == "crawler") {
		wikimode = "crawler";

		# 1. No queries
		if ((p = ENVIRON["QUERY_STRING"]) != ""  &&  p !~ /^op=(list|browse|main|crawler)$/) {
			sendredirect(BASEADR path, "", "");
			exit (0);
			}
		
		# 2. No options
		if (requestoptions > 0) {
			p = option["m"];
			delete option;
			BASEADR = makebaseurl(option, "m", p);
			sendredirect(BASEADR path, "", "");
			exit (0);
			}
		}


	# Find out which browser it is.

	if ((agent = tolower(ENVIRON["HTTP_USER_AGENT"])) ~ /opera/)
		isOpera = 1;
	else if (agent ~ /mozilla.*firefox/) {
		isFirefox = 1;
		applyfirefoxpatch = 1;
		wikiparser = wikiparser " -P";
		}
	else if (agent ~ /links|lynx|w3m/) {
		isPlain = 1;
		readonly = 1;
		}

	if ((p = option["m"]) == "pda"  ||  p == "plain")
		isPDA = 1;

	if (p == "pda"  ||  p == "sidebar")
		isSmall = 1;


	# Normal request processing starts here.

	filename = sprintf ("%s/%s", wikipath, path);
	wikidir = getdirname(filename);
	op = ENVIRON["CGI_OP"];
	if (path ~ /\/$/  &&  op ~ /^(list|browse|search)$/)
		;
	else if (path != ""  &&  stat(filename, sbuf) == 0  &&  sbuf[filename, "type"] == "dir") {
		if (path !~ /\/$/)
			sendredirect(p = BASEADR path "/");
		else
			sendredirect(p = BASEADR path wikiindex);

		exit (0);
		}

	if (send404 == 1  &&  sbuf[filename, "type"] == "."  &&  op != "edit"  &&
	    ENVIRON["REQUEST_METHOD"] != "POST") {
		send404resp(path);
		exit (0);
		}


	wikifile = filename;
#printf ("wikifile= %s, path= %s\n", wikifile, path) >>STDERR;
	if (path == "") {
		op = tolower(ENVIRON["CGI_OP"]);
		if (op == "") {
			sendredirect(BASEADR wikiindex, "operation not set");
			exit (0);
			}
		
		title = "00-WikiList";
		}


	subdirlocation = sprintf ("<A HREF=\"%s\">TOP</A>", BASEADR);
	if (wikisubdir != "/") {
		n = split(wikisubdir, x, "/");
		p = ""
		for (i=1; i<=n; i++) {
			if (x[i] == "")
				continue;

			p = p x[i] "/";
			subdirlocation = subdirlocation sprintf(" &gt; <A HREF=\"%s\">%s</A>",
						BASEADR p, x[i]);
			}
		}

	if (side != "") {		# Split Screen aktiv
#		if (wikiopt["lockframe"] == "unlocked") {
#			wikiopt["lockframe"] = "unlocked";
#			lockurl = sprintf ("%s/%s", ENVIRON["SCRIPT_NAME"], mkoptions(config, wikimode, side, "locked"));
#			}
#		else {
#			wikiopt["lockframe"] = "locked";
#			unlockurl = sprintf ("%s/%s", ENVIRON["SCRIPT_NAME"], mkoptions(config, wikimode, side));
#			}
#printf ("** lock= %s, unlock= %s\n", lockurl, unlockurl) >>STDERR;
		}


	fileurl = path;

	mime = "";
	if ((mime = getmimetype(path)) != "text/yawk") {
		if (mime == "text/yawk")
			;	# normal processing
		else if (mime == "text/plain"  &&  ENVIRON["CGI_OP"] != "")
			;	# pass-though
		else if (mime == "text/plain"  &&  wraptext == 1)
			;
		else if ((op = tolower(ENVIRON["CGI_OP"])) == "info"  ||  op == "rename")
			;
		else {
			sendheader(200, mime, 0);
			fflush();

			cmd = sprintf ("/bin/cat <%s", quote(filename));
			system(cmd);
			exit (0);
			}
		}

	if (ENVIRON["REQUEST_METHOD"] == "POST") {
		if (readonly != 0) {
			sendredirect(BASEADR "?op=list", "readonly mode, POST access denied");
			exit (0);
			}

		if (personalmode == 1) {
			if (ENVIRON["REMOTE_ADDR"] != "127.0.0.1") {
				sendredirect(BASEADR "?op=list", "personal mode, POST access denied");
				exit (0);
				}
			}

		username = ENVIRON["REMOTE_USER"];
		if (havewriteperms == 1) {
			if (writeperm[username] != 1) {
				sendredirect(BASEADR "?op=list", "user not permitted");
				exit (0);
				}
			}

		if (ENVIRON["CONTENT_TYPE"] ~ /^multipart\/form-data/) {
			cmd = sprintf ("%s/wiki-receiver '%s'", bindir, wikidir);
			while (cmd | getline > 0) {
				if ($2 == "file")
					filename = $3;
				}

			if (filename == "")
				sendredirect(BASEADR "?op=list", "no upload filename provided");
			else
				sendredirect(BASEADR filename "?op=info");

			exit (0);
			}

		if ((op = ENVIRON["CGI_OP"]) == "new") {
			redirect = sprintf ("%s%s", BASEADR, ENVIRON["CGI_NAME"]);
			sendredirect(redirect);
			}

#ifdef WITH_NUMBERED
		else if (op == "new-numbered") {
			file = sprintf ("%s/%s", wikipath, SERIALFILENAME);
			getline serialnum <file;
			close (file);

			serialnum = serialnum + 1;
			printf ("%06d\n", serialnum) >file;
			close (file);

			basename = sprintf ("%06d-%s", serialnum, SERIALWIKINAME);
			filename = sprintf ("%s/%s", wikidir, basename);
			if ((title = ENVIRON["CGI_TITLE"]) == "") {
#				title = sprintf ("%06d", serialnum);
				title = "Untitled Document";
				}

			data = sprintf("\n---* %06d - %s\n", serialnum, title);
			if ("CGI_TEXT" in ENVIRON)
				data = data "\n" noctrl(ENVIRON["CGI_TEXT"]) "\n";

			data = data sprintf ("\n[%06d]\n", serialnum);
			filetext[1] = data;

			writefile(filename, filedata, filetext, 1, "create");
			sendredir2(BASEADR basename, ENVIRON["CGI__REDIR"]);
			}
#endif

		else if (op == "store") {
#ifdef WITH_MIDAIRCOLL
			if (tolower(ENVIRON["CGI_SUBMIT"]) == "save") {
				datelastmod = sbuf[filename, "mtime"];

				if (side == ""  &&
				    ENVIRON["CGI_OVERWRITE"] == ""  &&
				    ENVIRON["CGI_DATELASTMOD"] != ""  &&
				    ENVIRON["CGI_DATELASTMOD"] != datelastmod) {
					sendheader(200, "text/html", 1);
					printf ("<B>Alert:</B> File changed since your started editing.<P>\n");
#					printf ("filename: %s<BR>\n", filename);
					printf ("last modified: %s<BR>\n", datelastmod);
					printf ("editing started: %s<P>\n", ENVIRON["CGI_DATELASTMOD"]);
					printf ("Use your browser's <I>Back</I> button and check the modifications.\n");
					printf ("You may also ignore the changes by setting the overwrite option and submitting");
					printf (" your text again.\n");
					exit (0);
					}
				}
#endif

			data = noctrl(ENVIRON["CGI_TEXT"]);
			gsub(/\r/, "", data);


#ifdef WITH_PREVIEW
			if (ENVIRON["CGI_PREVIEW"] != "") {
				sendheader(200, "text/html", 1);
				starthtml("Edit Page", cssadr, 0, "");
				navigator(BASEADR, "", 1);

				n = split(data, text, "\n");
				text[1] = "%previewnote=PREVIEW%=\n" text[1];

				filedata[""] = 1;
				if (ENVIRON["CGI__INPUTPREVIEW"] != "yes")
					sendtext("content", "", n, text);
				else {
					n = split(noctrl(ENVIRON["CGI_TEXT"]), text, "\n");
					sendtext("content", "", n, text);
					}

				if (ENVIRON["CGI_TARGET"] == "") {
					printf ("<HR SIZE=\"1\">\n");
					sub(/[^\n]*\n/, "", text[1]);
					__insert_editform(wikifile, n, text, ENVIRON["CGI_TARGET"]);
					}

				navigator(BASEADR, "", 0);
				endhtml();
				exit (0);
				}
#endif

			if (tolower(ENVIRON["CGI_SUBMIT"]) == "save") {
				printf ("%s\n", data) >filename;
				close (filename);
				}

			sendredirect(BASEADR path);
			}
		else if (op == "rename") {
			target = rewritename(ENVIRON["CGI_TARGET"], 0);
			sub(/^.*\//, "", target);
			copy = tolower(ENVIRON["CGI_COPY"]);
			if (target == "---") {
				cmd = sprintf("/bin/rm %s", quote(filename));
				system(cmd);
				path = wikisubdir "?op=list";
				}
			else if (copy == "yes") {
				copy = sprintf ("%s/%s", wikidir, target);
				while (getline line <filename > 0)
					printf ("%s\n", line) >copy;

				close(copy);
				close(file);
				path = wikisubdir target;
				}
			else if (target != "") {
				cmd = sprintf("/bin/mv %s %s",
					quote(filename), quote(wikidir "/" target));
				system(cmd);
				path = wikisubdir target;
				}

			if (getmimetype(path) != "text/yawk")
				sendredirect(BASEADR path "?op=info");
			else
				sendredirect(BASEADR path);
			}



#ifdef WITH_SUBDIR
		else if (op == "mkdir")
			makedir(wikisubdir, ENVIRON["CGI_NAME"]);
		else if (op == "rmdir")
			removedir(wikisubdir, ENVIRON["CGI_SUBDIR"]);
		else if (op == "movedir") {
			p = wikisubdir;
			sub(/\/+$/, "", p);
			sub(/^.*\//, "", p);
			dest = rewritename(ENVIRON["CGI_DEST"]) "/" p "/";
			renamedir(wikisubdir, dest);
			}
		else if (op == "renamedir") {
			name = rewritename(ENVIRON["CGI_NAME"]);
			sub(/^.*\//, "", name);
			newdir = wikisubdir;
			sub(/[^\/]*\/?$/, "", newdir);
			newdir = newdir name "/";
			renamedir(wikisubdir, newdir);
			}
#endif

		else if (op == "sync") {
			if (tolower(ENVIRON["CGI_SUBMIT"]) == "export") {
				title = "Wiki Export";
				cmd = exportcmd;
				}
			else {
				title = "Wiki Sync";
				cmd = synccmd;
				}

			if ((password = ENVIRON["CGI_PASSWD"]) != "") {
				gsub(/'/, "", password);
				cmd = cmd sprintf (" -p '%s'", password);
				}

			cmd = cmd sprintf (" '%s' 2>&1", wikipath);
			sendheader(200, "text/html", 0);
			starthtml(title, cssadr, 0, "");
			navigator(BASEADR, "", 1);

			printf ("<B>%s</B><P>\n", title);
			printf ("<PRE>\n");

			while (cmd | getline line > 0) {
				line = noctrl(escape(line, 0));
				printf ("%s\n", line);
				}

			close (cmd);
			printf ("</PRE>\n");

			navigator(BASEADR, "", 0);
			endhtml();
			}
		else
			sendredirect(BASEADR, "unknown operation: " op);
		}
	else if (ENVIRON["REQUEST_METHOD"] == "GET") {
		if (readonly != 0)
			;
		else if ((username = ENVIRON["REMOTE_USER"]) != "") {
			if (havewriteperms == 1  &&  writeperm[username] != 1)
				readonly = 1;
			}

		if (personalmode != 0  &&  ENVIRON["REMOTE_ADDR"] != "127.0.0.1")
			readonly = 1;

		op = ENVIRON["CGI_OP"];

		if (op == "xx-new") {
			sendheader(200, "text/html", 0);
			starthtml("Create New File", cssadr, 0, "");
			navigator(BASEADR, "", 1);

			printf ("<U>Create New File:</U><P>\n");
			printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
			printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"new\"></INPUT>\n");
			printf ("  <DL>\n");
			printf ("  <DT>Enter filename for the new file<DD>\n");
			printf ("  <INPUT TYPE=\"TEXT\" NAME=\"name\" VALUE=\"\" ACCESSKEY=\"N\" SIZE=\"30\">\n");
			printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Create\"></INPUT><BR>\n");
			printf ("  </DL>\n");
			printf ("</FORM>\n");

#ifdef WITH_NUMBERED
			printf ("<FORM ACTION=\"%s\" METHOD=\"POST\">\n", BASEADR);
			printf ("  <DL>\n");
			printf ("  <DT>or the file's title to get a new numbered file:<DD>\n");
			printf ("  <INPUT TYPE=\"HIDDEN\" NAME=\"op\" VALUE=\"new-numbered\">\n");
			printf ("  <INPUT TYPE=\"TEXT\" NAME=\"title\" VALUE=\"\" ACCESSKEY=\"U\" SIZE=\"30\">\n");
			printf ("  <INPUT TYPE=\"SUBMIT\" NAME=\"submit\" VALUE=\"Create\">\n");
			printf ("  </DL>\n");
			printf ("</FORM>\n");
#endif

			navigator(BASEADR, "", 0);
			endhtml();
			}
		else if (op == "new"  ||  op == "dirinfo") {
			sendheader(200, "text/html", 0);
			starthtml("Create New File", cssadr, 0, "");
			navigator(BASEADR, "", 1);

			printdirinfo(wikidir, wikisubdir);

			navigator(BASEADR, "", 0);
			endhtml();
			}
		else if (op == "search")
			dosearch(wikisubdir, path);
		else if (op == "open")
			doopen(wikisubdir);
		else if (op == "compound") {
			sendheader(200, "text/html", 0);
			readonly = 1;
			starthtml(title, css, 0, "text");
#			navigator(path, "", 1);
			docompound("compound", "", ENVIRON["CGI_S"]);
			endhtml();
			}
		else if (op == "frame") {
			l = (p = makebaseurl(option, "s", "left")) ((lm = ENVIRON["CGI_L"]) == ""? path: lm);
			r = makebaseurl(option, "s", "right") ((rm = ENVIRON["CGI_R"]) == ""? path: rm);
#printf ("l= %s, p= %s, lm= %s, path= %s\n", l, p, lm, path) >>STDERR;

			sendheader(200, "text/html", 1);
			starthtml(wikiname " [rl-Framed]", " ", 0, "");

			if (wikimode == "sidewiki"  ||  wikimode == "sidebar")
				printf ("<FRAMESET ROWS=\"50%c,*\" BORDER=\"1\" RESIZABLE>\n", "%");
			else
				printf ("<FRAMESET COLS=\"50%c,*\" BORDER=\"1\" RESIZABLE>\n", "%");

			printf ("  <FRAME NAME=\"left\" SRC=\"%s\">\n", l);
			printf ("  <FRAME NAME=\"right\" SRC=\"%s\">\n", r);

			printf ("</FRAMESET>\n");
			endhtml();
			}
		else if (op == "list")
			dolist(wikisubdir);
		else if (op == "browse")
			dobrowser(wikisubdir);
		else if (op == "tree") {
			sendheader(200, "text/html", 1);
			starthtml("Edit Page", cssadr, 0, "");
			navigator(BASEADR, "", 1);

			__printtree(wikipath, wikisubdir);
			navigator(BASEADR, "", 0);
			endhtml();
			}
		else if (op == "edit") {
			sendheader(200, "text/html", 1);
			starthtml("Edit Page", cssadr, 0, "");
			navigator(BASEADR, path, 1);

			n = readfile(wikifile, 0, text, 0, 0);
			__insert_editform(wikifile, n, text, ENVIRON["CGI_TARGET"]);

			navigator(BASEADR, path, 0);
			endhtml();
			}


		else if (op == "redir") {
			for (name in CGIVAR) {
				if (name == "CGI_OP"  ||  match(name, /^CGI_O/) == 0)
					continue;

#printf ("var= %s <%s>\n", name, ENVIRON["name"]) >>STDERR;
				c = tolower(substr(name, 6));
				option[c] = ENVIRON[name];
				rmvar(name);
#				p = sprintf ("(^|&)%s[^&]", tolower(substr(name, 5)));
#				gsub(p, "", ENVIRON["QUERY_STRING"]);
				}

			BASEADR = makebaseurl(option, "", "");
			if ((p = ENVIRON["CGI_D"]) != "")
				path = p;

#printf (">> url= %s\n", BASEADR path) >>STDERR;
			sendredirect(BASEADR path);
			}
		else if (op == "info") {
			sendheader(200, "text/html", 1);
			starthtml("Info: " path, "", 0, "");

			navigator(BASEADR, "", 1);
			__doinfopage(filename);

			endhtml();
			}

		else if (op == "as-binary") {
			sendheader(200, "application/octet-stream", 1);
			while (getline sline <filename > 0) {
				if (sline ~ /^:data:/)
					continue;

				printf ("%s%s", sline, RT);
				}
			}
		else if (op == "rawify") {
			options = "";
			if (wikiopt["lockframe"] != "")
				options = options sprintf ("%c%s", configdelim, wikiopt["lockframe"]);
			else
				wikiopt["lockframe"] = "unlocked";

			options = options sprintf ("%craw", configdelim);

			url = sprintf ("%s/-%s%c%s%s/", ENVIRON["SCRIPT_NAME"], config, configdelim, side, options);
			url = url wikisubdir path;
			sendredirect(url);
			}


		else if (path == "")
			sendredirect(BASEADR wikiindex);
		else {
			options = "";
			if (op == "raw")
				wikimode = "raw";

			if ((linenumber = ENVIRON["CGI_LINE"]+0) > 0)
				options = sprintf (" -l %s", quote(linenumber));

#printf ("wikifile= %s\n", wikifile) >>STDERR;
			n = readfile(wikifile, 0, text, 0, 0);

			sendheader(200, "text/html", 0);
			starthtml(title, cssadr, 0, path);
			navigator(BASEADR, path, 1);

#			printf ("<DIV CLASS=\"content\">\n");
			sendtext("content", options, n, text);

			if ("isindex" in fileconf)
				doindex(wikifile, "UL");

#			printf ("</DIV><P>\n");
			navigator(BASEADR, path, 0);
			endhtml();
			}
		}

	exit (0);
	}

